"""
Implementation of reaction network and graph classes.
"""
from __future__ import annotations

from dataclasses import field
from queue import Empty, PriorityQueue
from typing import TYPE_CHECKING, Iterable

import rustworkx as rx
from pymatgen.entries import Entry
from tqdm import tqdm

from rxn_network.costs.functions import Softplus
from rxn_network.entries.experimental import ExperimentalReferenceEntry
from rxn_network.network.base import Graph, Network
from rxn_network.network.entry import NetworkEntry, NetworkEntryType
from rxn_network.pathways.basic import BasicPathway
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger

if TYPE_CHECKING:
    from rxn_network.costs.base import CostFunction
    from rxn_network.reactions.base import Reaction

logger = get_logger(__name__)


class ReactionNetwork(Network):
    """
    Main reaction network class for building graph networks and performing
    pathfinding. Graphs are built using the rustworkx package (a NetworkX equivalent
    implemented in Rust).

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.

    """

    def __init__(
        self,
        rxns: ReactionSet,
        cost_function: CostFunction = field(default_factory=Softplus),
    ):
        """
        Initialize a ReactionNetwork object for a reaction set and cost function.

        To build the graph network, call the build() method in-place.

        Args:
            rxns: Set of reactions used to construct the network.
            cost_function: The function used to calculate the cost of each reaction
                edge. Defaults to a Softplus function with default settings (i.e.
                energy_per_atom only).
        """
        super().__init__(rxns=rxns, cost_function=cost_function)

    def build(self) -> None:
        """
        In-place method. Construct the reaction network graph object and store under the
        "graph" attribute.

        WARNING: This does NOT initialize the precursors or target attributes; you must
        call set_precursors() or set_target() to do so.

        Returns:
            None
        """
        logger.info("Building graph from reactions...")

        g = Graph()

        nodes, edges = get_rxn_nodes_and_edges(self.rxns)
        edges.extend(get_loopback_edges(nodes))  # type: ignore

        g.add_nodes_from(nodes)
        g.add_edges_from(edges)

        logger.info(f"Built graph with {g.num_nodes()} nodes and {g.num_edges()} edges")

        self._g = g  # type: ignore

    def find_pathways(
        self, targets: list[Entry | str], k: int = 15
    ) -> list[BasicPathway]:
        """
        Find the k-shortest paths to a provided list of one or more targets.

        Args:
            targets: List of the formulas or entry objects of each target.
            k: Number of k-shortest paths to find for each target. Defaults to 15.

        Returns:
            List of BasicPathway objects to all provided targets.
        """
        if not self.precursors:
            raise AttributeError("Must call set_precursors() before pathfinding!")

        paths = []
        for target in targets:
            self.set_target(target)
            print(
                f"Paths to {self.target.composition.reduced_formula} \n"  # type: ignore
            )
            print("--------------------------------------- \n")
            pathways = self._k_shortest_paths(k=k)
            paths.extend(pathways)

        paths = PathwaySet.from_paths(paths)

        return paths

    def set_precursors(self, precursors: Iterable[Entry | str]):
        """
        In-place method. Sets the precursors of the network. Removes all references to
        previous precursors.

        If entries are provided, will use the entries to set the precursors. If strings
        are provided, will automatically find the lowest-energy entries with matching
        reduced_formula.

        Args:
            precursors: iterable of entries/formulas of precursor phases.

        Returns:
            None
        """
        g = self._g
        if not g:
            raise ValueError("Must call build() before setting precursors!")

        precursors = {
            p
            if isinstance(p, (Entry, ExperimentalReferenceEntry))
            else self.entries.get_min_entry_by_formula(p)
            for p in precursors
        }

        if precursors == self.precursors:
            return
        if not all(p in self.entries for p in precursors):
            raise ValueError("One or more precursors are not included in network!")

        precursors_entry = NetworkEntry(precursors, NetworkEntryType.Precursors)
        if self.precursors:  # remove old precursors
            for node in g.node_indices():
                if (
                    g.get_node_data(node).description.value
                    == NetworkEntryType.Precursors.value
                ):
                    g.remove_node(node)
                    break
            else:
                raise ValueError("Old precursors node not found in graph!")

        precursors_node = g.add_node(precursors_entry)

        edges_to_add = []
        for node in g.node_indices():
            entry = g.get_node_data(node)
            entry_type = entry.description.value

            if entry_type == NetworkEntryType.Reactants.value:
                if entry.entries.issubset(precursors):
                    edges_to_add.append((precursors_node, node, "precursor_edge"))
            elif entry.description.value == NetworkEntryType.Products.value:
                for node2 in g.node_indices():
                    entry2 = g.get_node_data(node2)
                    if entry2.description.value == NetworkEntryType.Reactants.value:
                        if precursors.issuperset(entry2.entries):
                            continue
                        if precursors.union(entry.entries).issuperset(entry2.entries):
                            edges_to_add.append((node, node2, "loopback_edge"))

        g.add_edges_from(edges_to_add)
        self._precursors = precursors

    def set_target(self, target: Entry | str) -> None:
        """
        In-place method. Can only provide one target entry or formula at a time.

        If entry is provided, will use that entry to set the target. If string is
        provided, will automatically find minimum-energy entry with
        matching reduced_formula.

        Args:
            target: Entry, or string of reduced formula, of target phase.

        Returns:
            None
        """
        g = self._g
        if not g:
            raise ValueError("Must call build() before setting target!")

        target = (
            target
            if isinstance(target, (Entry, ExperimentalReferenceEntry))
            else self.entries.get_min_entry_by_formula(target)
        )

        if target == self.target:
            return

        if target not in self.entries:
            raise ValueError("Target is not included in network!")

        if self.target:
            for node in g.node_indices():
                if (
                    g.get_node_data(node).description.value
                    == NetworkEntryType.Target.value
                ):
                    g.remove_node(node)
                    break
            else:
                raise ValueError("Old target node not found in graph!")

        target_entry = NetworkEntry([target], NetworkEntryType.Target)
        target_node = g.add_node(target_entry)

        edges_to_add = []
        for node in g.node_indices():
            entry = g.get_node_data(node)
            entry_type = entry.description.value

            if entry_type != NetworkEntryType.Products.value:
                continue
            if target in entry.entries:
                edges_to_add.append((node, target_node, "target_edge"))

        g.add_edges_from(edges_to_add)

        self._target = target

    def _k_shortest_paths(self, k: int):
        """Wrapper for finding the k shortest paths using Yen's algorithm. Returns
        BasicPathway objects"""
        g = self._g
        if not g:
            raise ValueError("Must call build() before pathfinding!")
        paths = []

        precursors_node = g.find_node_by_weight(
            NetworkEntry(self.precursors, NetworkEntryType.Precursors)  # type: ignore
        )
        target_node = g.find_node_by_weight(
            NetworkEntry([self.target], NetworkEntryType.Target)
        )
        for path in yens_ksp(g, self.cost_function, k, precursors_node, target_node):
            paths.append(self._path_from_graph(g, path, self.cost_function))

        for path in paths:
            print(path, "\n")

        return paths

    @staticmethod
    def _path_from_graph(g, path, cf: CostFunction):
        """Gets a BasicPathway object from a shortest path found in the network"""
        rxns = []
        costs = []

        for step, node in enumerate(path):
            if (
                g.get_node_data(node).description.value
                == NetworkEntryType.Products.value
            ):
                e = g.get_edge_data(path[step - 1], node)

                rxns.append(e)
                costs.append(get_edge_weight(e, cf))

        return BasicPathway(reactions=rxns, costs=costs)


def get_rxn_nodes_and_edges(
    rxns: ReactionSet,
) -> tuple[list[NetworkEntry], list[tuple[int, int, Reaction]]]:
    """
    Given a reaction set, return a list of nodes and edges for constructing the
    reaction network.

    Args:
        rxns: a list of enumerated ComputedReaction objects to build a network from.

    Returns:
        A tuple consisting of (nodes, edges) where nodes is a list of NetworkEntry
        objects and edges is a list of tuples of the form (source_idx, target_idx,
        reaction).
    """
    nodes, edges = [], []

    for rxn in tqdm(rxns):
        reactant_node = NetworkEntry(rxn.reactant_entries, NetworkEntryType.Reactants)
        product_node = NetworkEntry(rxn.product_entries, NetworkEntryType.Products)

        if reactant_node not in nodes:
            nodes.append(reactant_node)
            reactant_idx = len(nodes) - 1
        else:
            reactant_idx = nodes.index(reactant_node)

        if product_node not in nodes:
            nodes.append(product_node)
            product_idx = len(nodes) - 1
        else:
            product_idx = nodes.index(product_node)

        edges.append((reactant_idx, product_idx, rxn))

    return nodes, edges


def get_loopback_edges(
    nodes: list[NetworkEntry],
) -> list[tuple[int, int, str]]:
    """
    Given a list of nodes to check, this function finds and returns loopback
    edges (i.e., edges that connect a product node to its equivalent reactant node)

    Args:
        nodes: List of vertices from which to find loopback edges

    Returns:
        A list of tuples of the form (source_idx, target_idx, reaction)
    """
    edges = []
    for idx1, p in enumerate(nodes):
        if p.description.value != NetworkEntryType.Products.value:
            continue
        for idx2, r in enumerate(nodes):
            if r.description.value != NetworkEntryType.Reactants.value:
                continue
            if p.entries == r.entries:
                edges.append((idx1, idx2, "loopback_edge"))

    return edges


def get_edge_weight(edge_obj: object, cf: CostFunction):
    """
    Given an edge of a reaction network, calculates the cost/weight of that edge.
    Corresponds to zero for loopback & precursor/target edges. Evaluates cost function
    for all reaction edges.

    Args:
        edge_obj: An edge in the reaction network
    """
    if isinstance(edge_obj, str) and edge_obj in [
        "loopback_edge",
        "precursor_edge",
        "target_edge",
    ]:
        return 0.0
    if isinstance(edge_obj, (ComputedReaction, OpenComputedReaction)):
        return cf.evaluate(edge_obj)

    raise ValueError("Unknown edge type")


def yens_ksp(
    g: rx.PyGraph,
    cf: CostFunction,
    num_k: int,
    precursors_node: int,
    target_node: int,
) -> list[list[int]]:
    """
    Yen's Algorithm for k-shortest paths, adopted for rustworkx.

    This implementation was inspired by the igraph implementation by Antonin Lenfant.

    Reference (original Yen's KSP paper):
        Jin Y. Yen, "Finding the K Shortest Loopless Paths n a Network", Management
        Science, Vol. 17, No. 11, Theory Series (Jul., 1971), pp. 712-716.

    Args:
        g: the rustworkx PyGraph object.
        cf: A cost function for evaluating the edge weights.
        num_k: number of k shortest paths that should be found.
        precursors_node: the index of the node representing the precursors.
        target_node: the index of the node representing the targets.

    Returns:
        List of lists of graph vertices corresponding to each shortest path
            (sorted in increasing order by cost).
    """

    def path_cost(nodes):
        """Calculates path cost given a list of nodes"""
        cost = 0
        for j in range(len(nodes) - 1):
            cost += get_edge_weight(g.get_edge_data(nodes[j], nodes[j + 1]), cf)
        return cost

    def get_edge_weight_with_cf(edge_obj):
        """Includes user-specified cost function in function call"""
        return get_edge_weight(edge_obj, cf)

    g = g.copy()

    path = rx.dijkstra_shortest_paths(  # type: ignore
        g, precursors_node, target_node, weight_fn=get_edge_weight_with_cf
    )

    if not path:
        return []

    path = list(path[target_node])

    a = [path]
    a_costs = [path_cost(path)]

    b = PriorityQueue()  # type: ignore

    for k in range(1, num_k):
        try:
            prev_path = a[k - 1]
        except IndexError:
            logger.info(f"Identified only k={k-1} paths before exiting. \n")
            break

        for i in range(len(prev_path) - 1):
            spur_node = prev_path[i]
            root_path = prev_path[:i]

            removed_edges = []

            for path in a:
                if len(path) - 1 > i and root_path == path[:i]:
                    try:
                        e = g.get_edge_data(path[i], path[i + 1])
                    except rx.NoEdgeBetweenNodes:
                        continue

                    g.remove_edge(path[i], path[i + 1])
                    removed_edges.append((path[i], path[i + 1], e))

            spur_path = rx.dijkstra_shortest_paths(  # type: ignore
                g, spur_node, target_node, weight_fn=get_edge_weight_with_cf
            )

            g.add_edges_from(removed_edges)

            if spur_path:
                total_path = list(root_path) + list(spur_path[target_node])
                total_path_cost = path_cost(total_path)
                b.put((total_path_cost, total_path))

        while True:
            try:
                cost_, path_ = b.get(block=False)
            except Empty:
                break
            if path_ not in a:
                a.append(path_)
                a_costs.append(cost_)
                break

    return a



"""
Experimental Gibbs free energy data from NIST-JANAF (compounds, gases), FREED (all
compounds), and FactSage (elemental chemical potentials)
"""
from pathlib import Path
from typing import Any, Dict, Union

from monty.serialization import loadfn

cwd = Path(__file__).parent.resolve()

PATH_TO_FREED = cwd / "freed"
PATH_TO_NIST = cwd / "nist"

G_ELEMS = loadfn(cwd / "mu_elements.json")


def load_experimental_data(fn: Union[str, Path]) -> Dict[str, Dict[float, Any]]:
    """
    Load experimental data from a json file.

    Args:
        fn: The filename of the json file
    """
    d = loadfn(fn)
    return {comp: make_float_keys(data) for comp, data in d.items()}


def make_float_keys(d) -> Dict[float, Any]:
    """
    Convert all keys in a dict to floats.
    """
    return {float(k): v for k, v in d.items()}



"""Utility functions used throughout the reaction-network package."""

import logging
import sys
from datetime import datetime
from itertools import chain, combinations, zip_longest
from pathlib import Path
from typing import Any, Iterable


def limited_powerset(iterable: Iterable, max_size: int) -> Iterable:
    """
    Helper method for generating subsets ranging from singular
    length to maximum length specified by max_size.

    Args:
        iterable: all objects to consider.
        max_size: upper limit for size of combination subsets.

    Returns:
        All combination sets up to maximum size
    """
    return chain.from_iterable(
        [combinations(iterable, num_combos) for num_combos in range(1, max_size + 1)]
    )


def grouper(iterable: Iterable, n: int, fillvalue: Any = None) -> Iterable:
    """
    Collects data into fixed-length chunks or blocks.

    Args:
        iterable: An iterable object to group.
        n: The number of items to include in each group.
        fillvalue: The value to use for the last group, if the length of the group is
            less than n.
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def get_project_root() -> Path:
    """
    Gets a Path object for the reaction-network project root directory.

    Note:
        This is specific to this file and project.

    Returns:
        Path object for the project root directory.
    """
    return Path(__file__).parent.parent.parent


def get_logger(
    name: str,
    level=logging.DEBUG,
    log_format="%(asctime)s %(levelname)s %(name)s %(message)s",
    stream=sys.stdout,
):
    """
    Code borrowed from the atomate package.

    Helper method for acquiring logger.
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)

    formatter = logging.Formatter(log_format)

    if logger.hasHandlers():
        logger.handlers.clear()

    sh = logging.StreamHandler(stream=stream)
    sh.setFormatter(formatter)

    logger.addHandler(sh)

    return logger


def datetime_str() -> str:
    """
    Get a string representation of the current time. Borrowed from atomate2 package.
    """
    return str(datetime.utcnow())



"""
Implements a class for storing (unbalanced/unconstrained) collection of reactions
forming a reaction pathway.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from rxn_network.pathways.base import Pathway

if TYPE_CHECKING:
    from rxn_network.reactions.base import Reaction


class BasicPathway(Pathway):
    """
    Simple pathway class for storing multiple ComputedReaction objects which form a
    single reaction pathway.

    In this class, there are no constraints on stoichiometry. See BalancedPathway for
    the more constrained version.
    """

    def __init__(self, reactions: list[Reaction], costs: list[float] | None = None):
        """
        Args:
            reactions: list of ComputedReaction objects making up the pathway.
            costs: Optional list of corresponding costs for each reaction.
        """
        self._reactions = reactions
        self.costs = costs or []

    @property
    def reactions(self) -> list[Reaction]:
        """A list of reactions contained in the reaction pathway"""
        return self._reactions

    @property
    def total_cost(self) -> float:
        """The sum of all costs associated with reactions in the pathway"""
        return sum(self.costs)

    @property
    def is_experimental(self) -> bool:
        """Whether or not all reactions in the pathway are experimental"""
        return all(e.is_experimental for e in self.entries)

    def __repr__(self) -> str:
        path_info = ""
        for rxn in self.reactions:
            path_info += f"{rxn} (dG = {round(rxn.energy_per_atom, 3)} eV/atom) \n"

        path_info += f"Total Cost: {round(self.total_cost,3)}"

        return path_info

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return all(
                other_rxn == rxn
                for other_rxn, rxn in zip(other.reactions, self.reactions)
            )

        return False

    def __hash__(self):
        return hash(tuple(self.reactions))



"""
This module implements two types of basic (combinatorial) reaction enumerators.
"""
from __future__ import annotations

from copy import deepcopy
from itertools import combinations, product
from math import comb
from typing import TYPE_CHECKING

import ray
from pymatgen.analysis.phase_diagram import GrandPotentialPhaseDiagram, PhaseDiagram
from tqdm import tqdm

from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.entries.utils import initialize_entry
from rxn_network.enumerators.base import Enumerator
from rxn_network.enumerators.utils import group_by_chemsys
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger, grouper, limited_powerset
from rxn_network.utils.ray import initialize_ray, to_iterator

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import ComputedEntry

logger = get_logger(__name__)


class BasicEnumerator(Enumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a maximum
    reactant/product cardinality (n); i.e., how many phases on either side of the
    reaction. This approach does not explicitly take into account thermodynamic
    stability (i.e. phase diagram). This allows for enumeration of reactions where the
    products may not be stable with respect to each other.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.
    """

    MIN_CHUNK_SIZE = 2500
    MAX_NUM_JOBS = 5000

    def __init__(
        self,
        precursors: list[str] | None = None,
        targets: list[str] | None = None,
        n: int = 2,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_duplicates: bool = False,
        filter_by_chemsys: str | None = None,
        chunk_size: int = MIN_CHUNK_SIZE,
        max_num_jobs: int = MAX_NUM_JOBS,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        max_num_constraints: int = 1,
        quiet: bool = False,
    ):
        """

        Args:
            precursors: Optional list of precursor formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as reactants. The "exclusive_precursors" parameter allows one to tune
                whether the enumerated reactions should include ALL precursors (the
                default) or just one.
            targets: Optional list of target formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as products. The "exclusive_targets" parameter allows one to tune
                whether the enumerated reactions should include ALL targets or just one
                (the default).
            n: Maximum reactant/product cardinality. This it the largest possible number
                of entries on either side of the reaction. Defaults to 2.
            exclusive_precursors: Whether enumerated reactions are required to have
                reactants that are a subset of the provided list of precursors. If True
                (the default), this only identifies reactions with reactants selected
                from the provided precursors.
            exclusive_targets: Whether enumerated reactions are required to have
                products that are a subset of the provided list of targets. If False,
                (the default), this identifies all reactions containing at least one
                composition from the provided list of targets (and any number of
                byproducts).
            filter_duplicates: Whether to remove duplicate reactions. Defaults to False.
            filter_by_chemsys: An optional chemical system for which to filter produced
                reactions by. This ensures that all output reactions contain at least
                one element within the provided system.
            chunk_size: The minimum number of reactions per chunk procssed. Needs to be
                sufficiently large to make parallelization a cost-effective strategy.
                Defaults to MIN_CHUNK_SIZE.
            max_num_jobs: The upper limit for the number of jobs created. Defaults to
                MAX_NUM_JOBS.
            remove_unbalanced: Whether to remove reactions which are unbalanced; this is
                usually advisable. Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. This is also
                advisable for ensuring that only unique reaction sets are produced.
                Defaults to True.
            max_num_constraints: The maximum number of allowable constraints enforced by
                reaction balancing. Defaults to 1 (which is usually advisable).
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """

        super().__init__(precursors=precursors, targets=targets)

        self.n = n
        self.exclusive_precursors = exclusive_precursors
        self.exclusive_targets = exclusive_targets
        self.filter_duplicates = filter_duplicates
        self.filter_by_chemsys = filter_by_chemsys
        self.chunk_size = chunk_size
        self.max_num_jobs = max_num_jobs
        self.remove_unbalanced = remove_unbalanced
        self.remove_changed = remove_changed
        self.max_num_constraints = max_num_constraints
        self.quiet = quiet

        self._stabilize = False

        self._p_set_func = "issuperset" if self.exclusive_precursors else "intersection"
        self._t_set_func = "issuperset" if self.exclusive_targets else "intersection"

        self.open_phases: list | None = None
        self._build_pd = False
        self._build_grand_pd = False

    def enumerate(
        self,
        entries: GibbsEntrySet,
    ) -> ReactionSet:
        """
        Calculate all possible reactions given a set of entries. If the enumerator was
        initialized with specified precursors or target, the reactions will be filtered
        by these constraints. Every enumerator follows a standard procedure:

        1) Initialize entries, i.e., ensure that precursors and target are considered
            stable entries within the entry set.

        2) Get a dictionary representing every possible "node", i.e. phase combination,
            grouped by chemical system.

        3) Filter the combos dictionary for chemical systems which are not relevant;
            i.e., don't contain elements in precursors and/or target.

        4) Iterate through each chemical system, initializing calculators,
            and computing all possible reactions for reactant/product pair and/or
            thermodynamically predicted reactions for given reactants.

        5) Add reactions to growing list, repeat Step 4 until combos dict exhausted.

        Args:
            entries: the set of all entries to enumerate from
        """
        if not ray.is_initialized():
            initialize_ray()

        entries, precursors, targets, open_entries = self._get_initialized_entries(
            entries
        )

        combos_dict = self._get_combos_dict(
            entries,
            precursors,
            targets,
            open_entries,
        )
        open_combos = self._get_open_combos(open_entries)

        if not open_combos:
            open_combos = []

        precursors = ray.put(precursors)
        targets = ray.put(targets)
        react_function = ray.put(self._react_function)
        open_entries = ray.put(open_entries)
        p_set_func = ray.put(self._p_set_func)
        t_set_func = ray.put(self._t_set_func)
        remove_unbalanced = ray.put(self.remove_unbalanced)
        remove_changed = ray.put(self.remove_changed)
        max_num_constraints = ray.put(self.max_num_constraints)

        entries_ref = ray.put(entries)

        num_cpus = int(ray.cluster_resources()["CPU"])

        pd_dict = {}
        if self._build_pd or self._build_grand_pd:
            # pre-loop for phase diagram construction
            pd_chunk_size = int(len(combos_dict) // num_cpus) + 1

            pd_dict_refs = []
            for item_chunk in grouper(combos_dict.items(), pd_chunk_size):
                pd_dict_refs.append(
                    _get_entries_and_pds.remote(
                        item_chunk,
                        entries_ref,
                        self.build_pd,
                        self.build_grand_pd,
                        getattr(self, "chempots", None),
                    )
                )

            for completed in tqdm(
                to_iterator(pd_dict_refs),
                total=len(pd_dict_refs),
                disable=self.quiet,
                desc=f"Building phase diagrams ({self.__class__.__name__})",
            ):
                pd_dict.update(completed)

        chunk_size = self.chunk_size
        total = sum(self._rxn_iter_length(c, open_combos) for c in combos_dict.values())

        if total / chunk_size > self.max_num_jobs:
            chunk_size = int(total // self.max_num_jobs) + 1
            logger.info(
                f"Increasing chunk size to {chunk_size} due to max job limit of"
                f" {self.max_num_jobs}"
            )

        to_run, current_chunk = [], []  # type: ignore
        for item in tqdm(
            combos_dict.items(),
            disable=self.quiet,
            desc="Building chunks...",
            total=len(combos_dict),
        ):
            chemsys, combos = item
            rxn_iter = list(self._get_rxn_iterable(combos, open_combos))

            filtered_entries, pd, grand_pd = None, None, None
            if self._build_pd or self._build_grand_pd:
                filtered_entries, pd, grand_pd = pd_dict[chemsys]

            filtered_entries = ray.put(filtered_entries)
            pd = ray.put(pd)
            grand_pd = ray.put(grand_pd)

            current_chunk_length = sum(len(c[0]) for c in current_chunk)  # type: ignore

            current_chunk.append(([], filtered_entries, pd, grand_pd))
            for r in rxn_iter:
                if current_chunk_length == chunk_size:
                    to_run.append(current_chunk)
                    current_chunk = [([r], filtered_entries, pd, grand_pd)]
                    current_chunk_length = 1
                else:
                    current_chunk[-1][0].append(r)
                    current_chunk_length += 1

            if current_chunk_length == chunk_size:
                to_run.append(current_chunk)
                current_chunk = []

        if current_chunk:
            to_run.append(current_chunk)

        rxn_chunk_refs, results = [], []  # type: ignore
        for chunk in to_run:
            rxn_chunk_refs.append(
                _react.remote(
                    chunk,
                    react_function,
                    open_entries,
                    precursors,
                    targets,
                    p_set_func,
                    t_set_func,
                    remove_unbalanced,
                    remove_changed,
                    max_num_constraints,
                )
            )

        for completed in tqdm(
            to_iterator(rxn_chunk_refs),
            total=len(to_run),
            disable=self.quiet,
            desc=f"Enumerating reactions ({self.__class__.__name__})",
        ):
            results.extend(completed)

        rxn_set = ReactionSet.from_rxns(
            results, entries=entries, filter_duplicates=self.filter_duplicates
        )

        return rxn_set

    @classmethod
    def _get_num_chunks(cls, items, open_combos, chunk_size) -> int:
        _ = open_combos  # not used in BasicEnumerator

        num_chunks = 0
        for _, i in items:
            num_combos = cls._rxn_iter_length(i, open_combos)
            num_chunks += num_combos // chunk_size + bool(num_combos % chunk_size)

        return num_chunks

    @staticmethod
    def _rxn_iter_length(combos, open_combos) -> int:
        _ = open_combos  # not used in BasicEnumerator
        return comb(len(combos), 2)

    def _get_combos_dict(
        self, entries, precursor_entries, target_entries, open_entries
    ):
        """
        Gets all possible entry combinations up to predefined cardinality (n), filtered
        and grouped by chemical system.
        """
        precursor_elems = [
            [str(el) for el in e.composition.elements] for e in precursor_entries
        ]
        target_elems = [
            [str(el) for el in e.composition.elements] for e in target_entries
        ]
        all_open_elems = {el for e in open_entries for el in e.composition.elements}

        entries = entries - open_entries

        combos = [set(c) for c in limited_powerset(entries, self.n)]
        combos_dict = group_by_chemsys(combos, all_open_elems)

        filtered_combos = self._filter_dict_by_elems(
            combos_dict,
            precursor_elems,
            target_elems,
            all_open_elems,
        )

        return filtered_combos

    def _get_open_combos(  # pylint: disable=useless-return
        self, open_entries
    ) -> list[set[ComputedEntry]] | None:
        """No open entries for BasicEnumerator, returns None"""
        _ = (self, open_entries)  # unused
        return None

    @staticmethod
    def _react_function(reactants, products, **kwargs):
        _ = kwargs  # unused
        forward_rxn = ComputedReaction.balance(reactants, products)
        backward_rxn = forward_rxn.reverse()
        return [forward_rxn, backward_rxn]

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Get all reaction/product combinations"""
        _ = open_combos  # unused

        return combinations(combos, 2)

    def _get_initialized_entries(self, entries):
        """Returns initialized entries, precursors, target, and open entries"""

        def initialize_entries_list(ents):
            new_ents = {initialize_entry(f, entries, self.stabilize) for f in ents}
            return new_ents

        precursors, targets = set(), set()

        entries_new = GibbsEntrySet(
            deepcopy(entries),
        )

        if self.precursors:
            precursors = initialize_entries_list(self.precursors)
        if self.targets:
            targets = initialize_entries_list(self.targets)

        for e in precursors | targets:
            if e not in entries_new:
                try:
                    old_e = entries_new.get_min_entry_by_formula(
                        e.composition.reduced_formula
                    )
                    entries_new.discard(old_e)
                except KeyError:
                    pass

                entries_new.add(e)

        if self.stabilize:
            entries_new = entries_new.filter_by_stability(e_above_hull=0.0)
            logger.info("Filtering by stable entries!")

        entries_new.build_indices()

        open_entries = set()
        if self.open_phases:
            open_entries = {
                e
                for e in entries_new
                if e.composition.reduced_formula in self.open_phases
            }

        return entries_new, precursors, targets, open_entries

    def _filter_dict_by_elems(
        self,
        combos_dict,
        precursor_elems,
        target_elems,
        all_open_elems,
    ):
        """Filters the dictionary of combinations by elements"""
        filtered_dict = {}

        all_precursor_elems = {el for g in precursor_elems for el in g}
        all_target_elems = {el for g in target_elems for el in g}
        all_open_elems = {str(el) for el in all_open_elems}

        filter_elems = None
        if self.filter_by_chemsys:
            filter_elems = set(self.filter_by_chemsys.split("-"))

        for chemsys, combos in combos_dict.items():
            elems = set(chemsys.split("-"))

            if filter_elems:
                if not elems.issuperset(filter_elems):
                    continue

            if len(elems) >= 10 or len(elems) == 1:  # too few or too many elements
                continue

            if precursor_elems:
                if not getattr(all_precursor_elems | all_open_elems, self._p_set_func)(
                    elems
                ):
                    continue

            if target_elems:
                if not getattr(all_target_elems | all_open_elems, self._t_set_func)(
                    elems
                ):
                    continue

            filtered_dict[chemsys] = combos

        return filtered_dict

    @property
    def stabilize(self) -> bool:
        """Whether or not to use only stable entries in analysis"""
        return self._stabilize

    @property
    def build_pd(self) -> bool:
        """Whether or not to build a PhaseDiagram object during reaction enumeration (
        useful for some analyses)"""
        return self._build_pd

    @property
    def build_grand_pd(self) -> bool:
        """Whether or not to build a GrandPotentialPhaseDiagram object during
        reaction enumeration (useful for some analyses)"""
        return self._build_grand_pd


class BasicOpenEnumerator(BasicEnumerator):
    """
    Enumerator for finding all simple reactions within a set of entries, up to a
    maximum reactant/product cardinality (n), with any number of open phases. Note:
    this does not return OpenComputedReaction objects (this can be calculated using
    the ReactionSet class).

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.
    """

    MIN_CHUNK_SIZE = 2500
    MAX_NUM_JOBS = 5000

    def __init__(
        self,
        open_phases: list[str],
        precursors: list[str] | None = None,
        targets: list[str] | None = None,
        n: int = 2,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_duplicates: bool = False,
        filter_by_chemsys: str | None = None,
        chunk_size: int = MIN_CHUNK_SIZE,
        max_num_jobs: int = MAX_NUM_JOBS,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        max_num_constraints: int = 1,
        quiet: bool = False,
    ):
        """
        Supplied target and calculator parameters are automatically initialized as
        objects during enumeration.

        Args:
            open_phases: List of formulas of open entries (e.g. ["O2"]).
            precursors: Optional list of precursor formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as reactants. The "exclusive_precursors" parameter allows one to tune
                whether the enumerated reactions should include ALL precursors (the
                default) or just one.
            targets: Optional list of target formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as products. The "exclusive_targets" parameter allows one to tune
                whether the enumerated reactions should include ALL targets or just one
                (the default).
            n: Maximum reactant/product cardinality. This it the largest possible number
                of entries on either side of the reaction. Defaults to 2.
            exclusive_precursors: Whether enumerated reactions are required to have
                reactants that are a subset of the provided list of precursors. If True
                (the default), this only identifies reactions with reactants selected
                from the provided precursors.
            exclusive_targets: Whether enumerated reactions are required to have
                products that are a subset of the provided list of targets. If False,
                (the default), this identifies all reactions containing at least one
                composition from the provided list of targets (and any number of
                byproducts).
            filter_duplicates: Whether to remove duplicate reactions. Defaults to False.
            filter_by_chemsys: An optional chemical system for which to filter produced
                reactions by. This ensures that all output reactions contain at least
                one element within the provided system.
            chunk_size: The minimum number of reactions per chunk procssed. Needs to be
                sufficiently large to make parallelization a cost-effective strategy.
                Defaults to MIN_CHUNK_SIZE.
            max_num_jobs: The upper limit for the number of jobs created. Defaults to
                MAX_NUM_JOBS.
            remove_unbalanced: Whether to remove reactions which are unbalanced; this is
                usually advisable. Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. This is also
                advisable for ensuring that only unique reaction sets are produced.
                Defaults to True.
            max_num_constraints: The maximum number of allowable constraints enforced by
                reaction balancing. Defaults to 1 (which is usually advisable).
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """
        super().__init__(
            precursors=precursors,
            targets=targets,
            n=n,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_duplicates=filter_duplicates,
            filter_by_chemsys=filter_by_chemsys,
            chunk_size=chunk_size,
            max_num_jobs=max_num_jobs,
            remove_unbalanced=remove_unbalanced,
            remove_changed=remove_changed,
            max_num_constraints=max_num_constraints,
            quiet=quiet,
        )
        self.open_phases: list[str] = open_phases

    @staticmethod
    def _rxn_iter_length(combos, open_combos):
        num_combos_with_open = sum(
            1 if not i & j else 0 for i in combos for j in open_combos
        )

        return len(combos) * num_combos_with_open

    def _get_open_combos(self, open_entries):
        """Get all possible combinations of open entries. For a single entry,
        this is just the entry itself."""
        open_combos = [
            set(c) for c in limited_powerset(open_entries, len(open_entries))
        ]
        return open_combos

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Get all reaction/product combinations."""
        combos_with_open = [
            combo | open_combo
            for combo in combos
            for open_combo in open_combos
            if not combo & open_combo
        ]
        rxn_iter = product(combos, combos_with_open)

        return rxn_iter


@ray.remote
def _react(
    chunk,
    react_function,
    open_entries,
    precursors,
    targets,
    p_set_func,
    t_set_func,
    remove_unbalanced,
    remove_changed,
    max_num_constraints,
):
    """
    This function is a wrapper for the specific react function of each enumerator. This
    wrapper contains the logic used for filtering out reactions based on the
    user-defined enumerator settings. It can also be called as a remote function using
    ray, allowing for parallel computation during reaction enumeration.

    WARNING: this function is not intended to to be called directly by the user and
    should only be used by the enumerator classes.

    """
    all_rxns = []
    for rxn_iterable, filtered_entries, pd, grand_pd in chunk:
        filtered_entries = ray.get(filtered_entries)
        pd = ray.get(pd)
        grand_pd = ray.get(grand_pd)

        for rp in rxn_iterable:
            if not rp:
                continue

            r = set(rp[0]) if rp[0] else set()
            p = set(rp[1]) if rp[1] else set()

            all_phases = r | p

            precursor_func = (
                getattr(precursors | open_entries, p_set_func)
                if precursors
                else lambda e: True
            )
            target_func = (
                getattr(targets | open_entries, t_set_func)
                if targets
                else lambda e: True
            )

            if (
                (r & p)
                or (precursors and not precursors & all_phases)
                or (p and targets and not targets & all_phases)
            ):
                continue

            if not (precursor_func(r) or (p and precursor_func(p))):
                continue
            if p and not (target_func(r) or target_func(p)):
                continue

            suggested_rxns = react_function(
                r, p, filtered_entries=filtered_entries, pd=pd, grand_pd=grand_pd
            )

            rxns = []
            for rxn in suggested_rxns:
                if (
                    rxn.is_identity
                    or (remove_unbalanced and not rxn.balanced)
                    or (remove_changed and rxn.lowest_num_errors != 0)
                    or rxn.data["num_constraints"] > max_num_constraints
                ):
                    continue

                reactant_entries = set(rxn.reactant_entries) - open_entries
                product_entries = set(rxn.product_entries) - open_entries

                if precursor_func(reactant_entries) and target_func(product_entries):
                    rxns.append(rxn)

            all_rxns.extend(rxns)

    return all_rxns


@ray.remote
def _get_entries_and_pds(
    combos_dict_chunk, entries, build_pd, build_grand_pd, chempots
):
    pd_dict = {}
    for item in combos_dict_chunk:
        if item is None:
            continue
        chemsys, _ = item
        elems = chemsys.split("-")

        filtered_entries = None
        pd = None
        grand_pd = None

        if build_pd or build_grand_pd:
            filtered_entries = entries.get_subset_in_chemsys(elems)

        if build_pd:
            pd = PhaseDiagram(filtered_entries)

        if build_grand_pd:
            grand_pd = GrandPotentialPhaseDiagram(filtered_entries, chempots)

        pd_dict[chemsys] = (filtered_entries, pd, grand_pd)
    return pd_dict



"""
Utility functions for plotting reaction data & performing analysis.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import plotly.express as px
import plotly.graph_objects as go
from pymatgen.analysis.chempot_diagram import plotly_layouts

from rxn_network.costs.pareto import get_pareto_front

if TYPE_CHECKING:
    from pandas import DataFrame


def plot_reaction_scatter(
    df: DataFrame,
    x: str = "secondary_competition",
    y: str = "energy",
    z: str | None = None,
    color: str | None = None,
    plot_pareto: bool = True,
) -> px.scatter:
    """
    Plot a Plotly scatter plot (2D or 3D) of reactions on various thermodynamic metric
    axes. This also constructs the Pareto front on the provided dimensions.

    Args:
        df: DataFrame with columns: rxn, energy, (primary_competition),
            (secondary_competition), (chempot_distance), (added_elems), (dE)

    Returns:
        Plotly scatter plot
    """

    def get_label_and_units(name):
        label, units = "", ""

        if name == "energy":
            label = (
                r"$\mathsf{Reaction~driving~force} ~"
                r"\mathrm{\left(\dfrac{\mathsf{eV}}{\mathsf{atom}}\right)}$"
            )
            units = "eV/atom"
            if z is not None:
                label = "Reaction Driving Force"
        elif name == "chempot_distance":
            label = (
                r"$\Sigma \Delta \mu_{\mathrm{min}} ~"
                r" \mathrm{\left(\dfrac{\mathsf{eV}}{\mathsf{atom}}\right)}$"
            )
            if z is not None:
                label = "Total chemical potential distance"
            units = "eV/atom"
        elif name == "primary_competition":
            label = "Primary Competition"
            units = "eV/atom"
        elif name == "secondary_competition":
            label = "Secondary Competition"
            units = "eV/atom"
        elif name == "dE":
            label = "Uncertainty"
            units = "eV/atom"

        return label, units

    df = df.copy()
    df["rxn"] = df["rxn"].astype(str)
    if "added_elems" in df:
        df["has_added_elems"] = df["added_elems"] != ""

    x_label, x_units = get_label_and_units(x)
    y_label, y_units = get_label_and_units(y)
    z_label, z_units = None, None

    cols: tuple = (x, y)

    if z is not None:
        z_label, z_units = get_label_and_units(z)
        cols = (x, y, z)

    if plot_pareto:
        pareto_df = get_pareto_front(df, metrics=cols)
        df = df.loc[~df.index.isin(pareto_df.index)]

        arr = pareto_df[list(cols)].to_numpy()
        if z is None:
            scatter = go.Scatter(
                x=arr[:, 0],
                y=arr[:, 1],
                hovertext=pareto_df["rxn"],
                marker={"size": 10, "color": "seagreen", "symbol": "diamond"},
                mode="markers",
                name="Pareto front",
            )
        else:
            scatter = go.Scatter3d(
                x=arr[:, 0],
                y=arr[:, 1],
                z=arr[:, 2],
                hovertext=pareto_df["rxn"],
                marker={"size": 10, "color": "seagreen", "symbol": "diamond"},
                mode="markers",
                name="Pareto front",
            )

    if z is None:
        layout_2d = plotly_layouts["default_layout_2d"]
        fig = px.scatter(
            df,
            x=x,
            y=y,
            hover_name="rxn",
            labels={x: x_label, y: y_label},
            color=color,
            color_discrete_map={True: "darkorange", False: "lightgray"},
        )
        fig.update_layout(layout_2d)
    else:
        layout_3d = plotly_layouts["default_layout_3d"]
        axis_layout = plotly_layouts["default_3d_axis_layout"].copy()
        axis_layout["titlefont"]["size"] = 14
        for t in ["xaxis", "yaxis", "zaxis"]:
            layout_3d["scene"][t] = axis_layout

        layout_3d["scene_camera"] = {
            "eye": {"x": -5, "y": -5, "z": 5},  # zoomed out
            "projection": {"type": "orthographic"},
            "center": {"x": -0.2, "y": -0.2, "z": -0.1},
        }

        fig = px.scatter_3d(
            df,
            x=x,
            y=y,
            z=z,
            hover_name="rxn",
            labels={x: x_label, y: y_label, z: z_label},
            template="simple_white",
            color=color,
            color_discrete_map={True: "darkorange", False: "lightgray"},
        )

        fig.update_layout(layout_3d)

    if plot_pareto:
        fig.add_trace(scatter)

    hovertemplate = (
        "<b>%{hovertext}</b><br>"
        + "<br><b>"
        + f"{x}"
        + "</b>: %{x:.3f}"
        + f" {x_units}"
        + "<br><b>"
        + f"{y}"
        + "</b>: %{y:.3f}"
        + f" {y_units}"
    )

    if z is not None:
        hovertemplate = (
            hovertemplate + "<br><b>" + f"{z}" + "</b>: %{z:.3f}" + f" {z_units}<br>"
        )

    fig.update_traces(hovertemplate=hovertemplate)

    return fig


def pretty_df_layout(df: DataFrame):
    """Improve visibility for a pandas DataFrame with wide column names"""
    return df.style.set_table_styles(
        [
            {
                "selector": "th",
                "props": [
                    ("max-width", "70px"),
                    ("text-overflow", "ellipsis"),
                    ("overflow", "hidden"),
                ],
            }
        ]
    )  # improve rendering in Jupyter


def filter_df_by_precursors(df: DataFrame, precursors: list[str]):
    """Filter a reaction DataFrame by available precursors"""
    df = df.copy()
    df["precursors"] = [
        list(sorted([r.reduced_formula for r in rxn.reactants])) for rxn in df["rxn"]
    ]
    selected = df[df["precursors"].apply(lambda x: all(p in precursors for p in x))]
    return selected.drop(columns=["precursors"])



"""
Basic interface and implementation for a Calculator and CostFunction.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING

from monty.json import MSONable

if TYPE_CHECKING:
    from rxn_network.reactions.base import Reaction


class Calculator(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction property calculator.
    """

    @abstractmethod
    def calculate(self, rxn: Reaction) -> float:
        """
        Evaluates a particular property of a reaction relevant to its cost ranking.
        """

    def calculate_many(self, rxns: list[Reaction]) -> list[float]:
        """
        Convenience method for performing calculate() on a list of reactions.

        Args:
            rxns: the list of Reaction objects to be evaluated

        Returns:
            A list of the reactions' calculated property values.
        """
        results = []
        for rxn in rxns:
            results.append(self.calculate(rxn))

        return results

    def decorate(self, rxn: Reaction) -> Reaction:
        """
        Returns a copy of the reaction with the calculated property by
        storing the value within the reaction's data dictionary.

        Args:
            rxn: The reaction object.

        Returns:
            A deep copy of the original reaction with a modified data dict.
        """
        new_rxn = deepcopy(rxn)

        if not getattr(new_rxn, "data"):
            new_rxn.data = {}

        new_rxn.data[self.name] = self.calculate(new_rxn)
        return new_rxn

    def decorate_many(self, rxns: list[Reaction]) -> list[Reaction]:
        """
        Convenience method for performing decorate() on a list of reactions.

        Args:
            rxns: the list of Reaction objects to be decorated

        Returns:
            A list of new (copied) reactions with modified data containing their
            calculated properties.
        """
        new_rxns = []
        for rxn in rxns:
            new_rxns.append(self.decorate(rxn))
        return new_rxns


class CostFunction(MSONable, metaclass=ABCMeta):
    """
    Base definition for a cost function.
    """

    @abstractmethod
    def evaluate(self, rxn: Reaction) -> float:
        """
        Evaluates the specified cost function equation on a reaction object.
        """



"""
This module for defining chemical reaction objects was originally sourced from
pymatgen and streamlined for the reaction-network code.
"""
from __future__ import annotations

import re
from copy import deepcopy
from functools import cached_property
from itertools import chain, combinations
from typing import TYPE_CHECKING

import numpy as np
from monty.fractions import gcd_float

from rxn_network.core import Composition
from rxn_network.reactions.base import Reaction

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element

TOLERANCE = 1e-6  # Tolerance for determining if a particular component fraction is > 0.


class BasicReaction(Reaction):
    """
    An object representing a basic chemical reaction: compositions and their
    coefficients.
    """

    def __init__(
        self,
        compositions: list[Composition],
        coefficients: list[float] | np.ndarray,
        balanced: bool | None = None,
        data: dict | None = None,
        lowest_num_errors: int | float = 0,
    ):
        """
        A BasicReaction object is defined by a list of compositions and their
        corresponding coefficients, where a negative coefficient refers to a
        reactant, and a positive coefficient refers to a product.

        Args:
            compositions: List of composition objects (pymatgen).
            coefficients: List of coefficients, where negative coeff distinguishes a
                reactant.
            balanced: Whether the reaction is stoichiometricaly balanced or not
                (see construction via balance() method).
            data: Optional corresponding data in dictionary format; often used to store
                various calculated parameters.
            lowest_num_errors: the minimum number of errors reported by the reaction
                balancing algorithm (see the balance() method). A number of errors
                >= 1 means that the reaction may be different than intended (some
                phases may be shuffled or removed entirely).
        """
        self._compositions = [Composition(c) for c in compositions]
        self._coefficients = np.array(coefficients)

        self.reactant_coeffs = {
            comp: coeff
            for comp, coeff in zip(self._compositions, self._coefficients)
            if coeff < 0
        }
        self.product_coeffs = {
            comp: coeff
            for comp, coeff in zip(self._compositions, self._coefficients)
            if coeff > 0
        }

        if balanced is not None:
            self.balanced = balanced
        else:
            sum_reactants = sum(
                [k * abs(v) for k, v in self.reactant_coeffs.items()], Composition({})
            )
            sum_products = sum(
                [k * abs(v) for k, v in self.product_coeffs.items()], Composition({})
            )

            if not sum_reactants.almost_equals(sum_products, rtol=0, atol=TOLERANCE):
                self.balanced = False
            else:
                self.balanced = True

        self.data = data if data else {}
        self.lowest_num_errors = lowest_num_errors

    @classmethod
    def balance(
        cls,
        reactants: list[Composition],
        products: list[Composition],
        data: dict | None = None,
    ) -> BasicReaction:
        """
        Reactants and products to be specified as list of
        pymatgen.core.Composition. e.g., [comp1, comp2]

        Args:
            reactants: List of reactants.
            products: List of products.
            data: Optional dictionary containing extra data about the reaction.
        """
        compositions = list(reactants + products)
        coeffs, lowest_num_errors, num_constraints = cls._balance_coeffs(
            reactants, products
        )
        if not data:
            data = {}
        data["num_constraints"] = num_constraints

        balanced = True
        if coeffs is None or lowest_num_errors == np.inf:
            balanced = False
            coeffs = np.zeros(len(compositions))

        return cls(
            compositions=compositions,
            coefficients=coeffs,
            balanced=balanced,
            data=data,
            lowest_num_errors=lowest_num_errors,
        )

    def normalize_to(self, comp: Composition, factor: float = 1) -> BasicReaction:
        """
        Normalizes the reaction to one of the compositions via the provided factor.

        By default, normalizes such that the composition given has a coefficient of
        1.

        Args:
            comp: Composition object to normalize to
            factor: factor to normalize to. Defaults to 1.
        """
        all_comp = self.compositions
        coeffs = self.coefficients.copy()
        scale_factor = abs(1 / coeffs[self.compositions.index(comp)] * factor)
        coeffs *= scale_factor
        return BasicReaction(all_comp, coeffs)

    def normalize_to_element(
        self, element: Element, factor: float = 1
    ) -> BasicReaction:
        """
        Normalizes the reaction to one of the elements.
        By default, normalizes such that the amount of the element is 1.
        Another factor can be specified.

        Args:
            element (Element/Species): Element to normalize to.
            factor (float): Factor to normalize to. Defaults to 1.
        """
        all_comp = self.compositions
        coeffs = self.coefficients.copy()
        current_el_amount = (
            sum(all_comp[i][element] * abs(coeffs[i]) for i in range(len(all_comp))) / 2
        )
        scale_factor = factor / current_el_amount
        coeffs *= scale_factor
        return BasicReaction(all_comp, coeffs)

    def get_el_amount(self, element: Element) -> float:
        """
        Returns the amount of the element in the reaction.

        Args:
            element (Element/Species): Element in the reaction

        Returns:
            Amount of that element in the reaction.
        """
        return (
            sum(
                self.compositions[i][element] * abs(self.coefficients[i])
                for i in range(len(self.compositions))
            )
            / 2
        )

    def get_coeff(self, comp: Composition):
        """
        Returns coefficient for a particular composition
        """
        return self.coefficients[self.compositions.index(comp)]

    def normalized_repr_and_factor(self):
        """
        Normalized representation for a reaction
        For example, ``4 Li + 2 O -> 2Li2O`` becomes ``2 Li + O -> Li2O``
        """
        return self._str_from_comp(self.coefficients, self.compositions, True)

    def copy(self) -> BasicReaction:
        """Returns a copy of the BasicReaction object"""
        return BasicReaction(
            compositions=self.compositions,
            coefficients=self.coefficients.copy(),
            balanced=self.balanced,
            data=self.data,
            lowest_num_errors=self.lowest_num_errors,
        )

    def reverse(self) -> BasicReaction:
        """
        Returns a copy of the original BasicReaction object where original reactants are
        new products, and vice versa.
        """
        return BasicReaction(
            compositions=self.compositions,
            coefficients=-1 * self.coefficients.copy(),
            balanced=self.balanced,
            data=self.data,
            lowest_num_errors=self.lowest_num_errors,
        )

    def is_separable(self, target: str | Composition) -> bool:
        """
        Checks if the reaction forms byproducts which are separable from the target
        composition; i.e., byproducts do not contain any of the elements in the target
        phase.

        Args:
            target: Composition of target; elements in this phase will be used to
                determine whether byproducts only contain added elements.

        Returns:
            True if reaction is separable from target, False otherwise.
        """
        target = Composition(target)
        identified_targets = [
            c
            for c in self.compositions
            if c.reduced_composition == target.reduced_composition
        ]

        if len(identified_targets) == 0:
            raise ValueError(f"Target composition {target} not in reaction {self}")

        added_elems = set(self.elements) - set(target.elements)
        products = set(deepcopy(self.products))

        for t in identified_targets:
            products.remove(t)

        separable = [added_elems.issuperset(comp.elements) for comp in products]
        found = all(separable)

        return found

    @cached_property
    def reactant_atomic_fractions(self) -> dict:
        """
        Returns the atomic mixing ratio of reactants in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        return {
            c.reduced_composition: -coeff * c.num_atoms / self.num_atoms
            for c, coeff in self.reactant_coeffs.items()
        }

    @cached_property
    def product_atomic_fractions(self) -> dict:
        """
        Returns the atomic mixing ratio of reactants in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        return {
            c.reduced_composition: coeff * c.num_atoms / self.num_atoms
            for c, coeff in self.product_coeffs.items()
        }

    @cached_property
    def reactant_molar_fractions(self) -> dict:
        """
        Returns the molar mixing ratio of reactants in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        total = sum(self.reactant_coeffs.values())

        return {c: coeff / total for c, coeff in self.reactant_coeffs.items()}

    @cached_property
    def product_molar_fractions(self) -> dict:
        """
        Returns the molar mixing ratio of products in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        total = sum(self.product_coeffs.values())

        return {c: coeff / total for c, coeff in self.product_coeffs.items()}

    @classmethod
    def from_string(cls, rxn_string: str) -> BasicReaction:
        """
        Generates a balanced reaction from a string. The reaction must
        already be balanced.

        Args:
            rxn_string:
                The reaction string. For example, "4 Li + O2-> 2 Li2O"

        Returns:
            BalancedReaction
        """
        rct_str, prod_str = rxn_string.split("->")

        def get_comp_amt(comp_str):
            return {
                Composition(m.group(2)): float(m.group(1) or 1)
                for m in re.finditer(
                    r"([\d\.]*(?:[eE]-?[\d\.]+)?)\s*([A-Z][\w\.\(\)]*)", comp_str
                )
            }

        reactant_coeffs = get_comp_amt(rct_str)
        product_coeffs = get_comp_amt(prod_str)

        return cls._from_coeff_dicts(reactant_coeffs, product_coeffs)

    @classmethod
    def from_formulas(cls, reactants: list[str], products: list[str]) -> BasicReaction:
        """
        Initialize a reaction from a list of 1) reactant formulas and 2) product
        formulas.

        Args:
            reactants: List of reactant formulas
            products: List of product formulas

        Returns:
            A BasicReaction object
        """

        reactant_comps = [Composition(r) for r in reactants]
        product_comps = [Composition(p) for p in products]
        rxn = cls.balance(reactants=reactant_comps, products=product_comps)

        return rxn

    @property
    def reactants(self) -> list[Composition]:
        """List of reactants for this reaction"""
        return list(self.reactant_coeffs.keys())

    @property
    def products(self) -> list[Composition]:
        """List of products for this reaction"""
        return list(self.product_coeffs.keys())

    @property
    def compositions(self) -> list[Composition]:
        """List of composition objects for this reaction"""
        return self._compositions

    @property
    def coefficients(self) -> np.ndarray:
        """Array of reaction coefficients"""
        return self._coefficients

    @cached_property
    def num_atoms(self) -> float:
        """Total number of atoms in this reaction"""
        return sum(
            coeff * sum(c[el] for el in self.elements)
            for c, coeff in self.product_coeffs.items()
        )

    @cached_property
    def energy(self) -> float:
        """The energy of this reaction"""
        raise ValueError("No energy for a basic reaction!")

    @cached_property
    def energy_per_atom(self) -> float:
        """The energy per atom of this reaction"""
        raise ValueError("No energy per atom for a basic reaction!")

    @cached_property
    def is_identity(self) -> bool:
        """Returns True if the reaction has identical reactants and products"""
        return self._get_is_identity()

    def _get_is_identity(self):
        """Returns True if the reaction has identical reactants and products"""
        if set(self.reactants) != set(self.products):
            return False
        if self.balanced is False:  # if not balanced, can not check coefficients
            return True
        return all(
            np.isclose(self.reactant_coeffs[c] * -1, self.product_coeffs[c])
            for c in self.reactant_coeffs
        )

    @cached_property
    def chemical_system(self) -> str:
        """Returns the chemical system as string in the form of A-B-C-..."""
        return "-".join(sorted([str(el) for el in self.elements]))

    @property
    def normalized_repr(self) -> str:
        """
        A normalized representation of the reaction. All factors are converted
        to lowest common factors.
        """
        return self.normalized_repr_and_factor()[0]

    @classmethod
    def _balance_coeffs(
        cls, reactants: list[Composition], products: list[Composition]
    ) -> tuple[np.ndarray, int | float, int]:
        """
        Balances the reaction and returns the new coefficient matrix
        """
        compositions = reactants + products
        num_comp = len(compositions)

        all_elems = sorted({elem for c in compositions for elem in c.elements})
        num_elems = len(all_elems)

        comp_matrix = np.array([[c[el] for el in all_elems] for c in compositions]).T

        rank = np.linalg.matrix_rank(comp_matrix)
        diff = num_comp - rank
        num_constraints = diff if diff >= 2 else 1

        # an error = a component changing sides or disappearing
        lowest_num_errors = np.inf

        first_product_idx = len(reactants)

        # start with simplest product constraints, work to more complex constraints
        product_constraints = chain.from_iterable(
            [
                combinations(range(first_product_idx, num_comp), n_constr)
                for n_constr in range(num_constraints, 0, -1)
            ]
        )
        reactant_constraints = chain.from_iterable(
            [
                combinations(range(0, first_product_idx), n_constr)
                for n_constr in range(num_constraints, 0, -1)
            ]
        )
        best_soln = np.zeros(num_comp)

        for constraints in chain(product_constraints, reactant_constraints):
            n_constr = len(constraints)

            comp_and_constraints = np.append(
                comp_matrix, np.zeros((n_constr, num_comp)), axis=0
            )
            b = np.zeros((num_elems + n_constr, 1))
            b[-n_constr:] = 1 if min(constraints) >= first_product_idx else -1

            for num, idx in enumerate(constraints):
                comp_and_constraints[num_elems + num, idx] = 1
                # arbitrarily fix coeff to 1

            coeffs = np.matmul(np.linalg.pinv(comp_and_constraints), b)

            num_errors = 0
            if np.allclose(np.matmul(comp_matrix, coeffs), np.zeros((num_elems, 1))):
                expected_signs = np.array([-1] * len(reactants) + [+1] * len(products))
                num_errors = np.sum(np.multiply(expected_signs, coeffs.T) < TOLERANCE)
                if num_errors == 0:
                    lowest_num_errors = 0
                    best_soln = coeffs
                    break
                if num_errors < lowest_num_errors:
                    lowest_num_errors = num_errors
                    best_soln = coeffs

        return np.squeeze(best_soln), lowest_num_errors, num_constraints

    @staticmethod
    def _from_coeff_dicts(reactant_coeffs, product_coeffs) -> BasicReaction:
        reactant_comps, r_coefs = zip(
            *[(comp, -1 * coeff) for comp, coeff in reactant_coeffs.items()]
        )
        product_comps, p_coefs = zip(*list(product_coeffs.items()))
        return BasicReaction(reactant_comps + product_comps, r_coefs + p_coefs)

    @staticmethod
    def _str_from_formulas(coeffs, formulas, tol=TOLERANCE) -> str:
        reactant_str = []
        product_str = []
        for amt, formula in zip(coeffs, formulas):
            if abs(amt + 1) < tol:
                reactant_str.append(formula)
            elif abs(amt - 1) < tol:
                product_str.append(formula)
            elif amt < -tol:
                reactant_str.append(f"{-amt:.4g} {formula}")
            elif amt > tol:
                product_str.append(f"{amt:.4g} {formula}")

        return " + ".join(reactant_str) + " -> " + " + ".join(product_str)

    @classmethod
    def _str_from_comp(cls, coeffs, compositions, reduce=False) -> tuple[str, float]:
        r_coeffs = np.zeros(len(coeffs))
        r_formulas = []
        for i, (amt, comp) in enumerate(zip(coeffs, compositions)):
            formula, factor = comp.get_reduced_formula_and_factor()
            r_coeffs[i] = amt * factor
            r_formulas.append(formula)
        if reduce:
            factor = 1 / gcd_float(np.abs(r_coeffs))
            r_coeffs *= factor
        else:
            factor = 1
        return cls._str_from_formulas(r_coeffs, r_formulas), factor

    def __eq__(self, other) -> bool:
        if self is other:
            return True

        if not self.chemical_system == other.chemical_system:
            return False

        if not len(self.products) == len(other.products):
            return False

        if not len(self.reactants) == len(other.reactants):
            return False

        if not np.allclose(sorted(self.coefficients), sorted(other.coefficients)):
            return False

        if not set(self.reactants) == set(other.reactants):
            return False

        if not set(self.products) == set(other.products):
            return False

        return True

    def __hash__(self):
        return hash(
            (self.chemical_system, tuple(sorted(self.coefficients)))
        )  # not checking here for reactions that are multiples (too expensive)

    def __str__(self) -> str:
        return self._str_from_comp(self.coefficients, self.compositions)[0]

    def __repr__(self) -> str:
        return self.__str__()



"""
Functions for visualizing/plotting reaction networks.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.cm
import numpy as np
from rustworkx.visualization import mpl_draw

if TYPE_CHECKING:
    from rustworkx import PyGraph


def plot_network(graph: PyGraph, vertex_cmap_name: str = "jet", **kwargs):
    """
    Plots a reaction network using rustworkx visualization tools (i.e., mpl_draw).

    Args:
        graph: a rustworkx PyGraph object
        vertex_cmap_name: the name of . Defaults to "jet".
        **kwargs: keyword arguments to pass to mpl_draw

    """
    g = graph.copy()

    node_names = [e.chemsys for e in g.nodes()]
    color_func_v = _get_cmap_string(vertex_cmap_name, domain=sorted(node_names))
    vertex_colors = [color_func_v(chemsys) for chemsys in node_names]

    return mpl_draw(
        g,
        node_size=2,
        width=0.2,
        arrow_size=3,
        node_color=vertex_colors,
        alpha=0.8,
        **kwargs,
    )


def _get_cmap_string(palette, domain):
    """
    Utility function for getting a matplotlib colormap string for a given palette and
    domain.
    """
    domain_unique = np.unique(domain)
    hash_table = {key: i_str for i_str, key in enumerate(domain_unique)}
    mpl_cmap = matplotlib.cm.get_cmap(palette, lut=len(domain_unique))

    def cmap_out(X, **kwargs):
        return mpl_cmap(hash_table[X], **kwargs)

    return cmap_out



"""
Entry objects used in a Network. These network entry objects hold multiple entries and
can be used as data for a node in the graph.
"""
from __future__ import annotations

from enum import Enum, auto
from typing import TYPE_CHECKING, Collection

from monty.json import MSONable
from monty.serialization import MontyDecoder

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element
    from pymatgen.entries import Entry


class NetworkEntryType(Enum):
    """Describes the Network Entry Type"""

    Precursors = auto()
    Reactants = auto()
    Products = auto()
    Target = auto()
    Dummy = auto()


class NetworkEntry(MSONable):
    """
    Helper class for describing combinations of ComputedEntry-like objects in context
    of a reaction network. This entry will represent a node in the network.
    """

    def __init__(self, entries: Collection[Entry], description: NetworkEntryType):
        """
        Args:
            entries: Collection of Entry-like objects
            description: Node type (e.g., Precursors, Target... see NetworkEntryType
                class)
        """
        self._entries = set(entries)
        self._elements = sorted(
            list({elem for entry in entries for elem in entry.composition.elements})
        )
        self._chemsys = "-".join([str(e) for e in self.elements])
        self._dim = len(self.chemsys)
        self._description = description

    @property
    def entries(self) -> set[Entry]:
        return self._entries

    @property
    def elements(self) -> list[Element]:
        return self._elements

    @property
    def chemsys(self) -> str:
        return self._chemsys

    @property
    def dim(self) -> int:
        return self._dim

    @property
    def description(self) -> NetworkEntryType:
        return self._description

    def as_dict(self) -> dict:
        """MSONable dict representation"""
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "entries": list(self.entries),
            "description": self.description.value,
        }

    @classmethod
    def from_dict(cls, d: dict) -> NetworkEntryType:
        """Load from MSONable dict"""
        return cls(
            MontyDecoder().process_decoded(d["entries"]),
            NetworkEntryType(d["description"]),
        )

    def __repr__(self) -> str:
        formulas = [entry.composition.reduced_formula for entry in self.entries]
        formulas.sort()
        return f"{self.description.name}: {','.join(formulas)}"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            if self.description == other.description:
                if self.chemsys == other.chemsys:
                    return self.entries == other.entries
        return False

    def __hash__(self):
        return hash((self.description, frozenset(self.entries)))


class DummyEntry(NetworkEntry):
    """
    A Dummy Entry that doesn't hold any info. This maybe useful for serving as an empty
    node to facilitate pathfinding to all nodes, etc.
    """

    def __init__(self):
        """Dummy node doesn't need any parameters"""
        self._entries = set()
        self._elements = []
        self._chemsys = ""
        self._dim = 0
        self._description = NetworkEntryType.Dummy

    def __repr__(self) -> str:
        return "Dummy Node"

    def __eq__(self, other) -> bool:
        return self is other

    def __hash__(self):
        return hash("Dummy")



"""
This module implements added features to the ChemicalPotentialDiagram class from
pymatgen.
"""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.analysis.chempot_diagram import ChemicalPotentialDiagram as ChempotDiagram
from pymatgen.analysis.phase_diagram import PhaseDiagram
from scipy.spatial import HalfspaceIntersection, KDTree

from rxn_network.entries.entry_set import GibbsEntrySet

if TYPE_CHECKING:
    from pymatgen.analysis.phase_diagram import PDEntry
    from pymatgen.core.periodic_table import Element


class ChemicalPotentialDiagram(ChempotDiagram):
    """
    This class is an extension of the ChemicalPotentialDiagram class from pymatgen.
    Several features have been added to the original class for the purpose of
    efficiently calculating the shortest distance between two chemical potential
    domains.

    For more information on this specific implementation of the algorithm, please
    cite/reference the paper below:

        Todd, P. K., McDermott, M. J., Rom, C. L., Corrao, A. A., Denney, J. J.,
        Dwaraknath, S. S.,  Khalifah, P. G., Persson, K. A., & Neilson, J. R. (2021).
        Selectivity in Yttrium Manganese Oxide Synthesis via Local Chemical Potentials
        in Hyperdimensional Phase Space. Journal of the American Chemical Society,
        143(37), 15185-15194. https://doi.org/10.1021/jacs.1c06229
    """

    def __init__(  # pylint: disable=super-init-not-called
        self,
        entries: list[PDEntry],
        limits: dict[Element, float] | None = None,
        default_min_limit: float | None = -100.0,
    ):
        """
        Initialize a ChemicalPotentialDiagram object.

        Args:
            entries: List of PDEntry-like objects containing a composition and
                energy. Must contain elemental references and be suitable for typical
                phase diagram construction. Entries must be within a chemical system
                with 2 or more elements.
            limits: Bounds of elemental chemical potentials (min, max), which are
                used to construct the border hyperplanes used in the
                HalfspaceIntersection algorithm; these constrain the space over which
                the domains are calculated and also determine the size of the plotted
                diagram. Any elemental limits not specified are covered in the
                default_min_limit argument.
            default_min_limit (float): Default minimum chemical potential limit for
                unspecified elements within the "limits" argument. This results in
                default limits of (-100, 0).
        """
        self.entries = list(
            sorted(entries, key=lambda e: e.composition.reduced_formula)
        )
        self._entry_set = GibbsEntrySet(self.entries)
        self.limits = limits
        self.default_min_limit = default_min_limit
        self.elements = list(
            sorted({els for e in self.entries for els in e.composition.elements})
        )
        self.dim = len(self.elements)
        self._min_entries, self._el_refs = self._get_min_entries_and_el_refs(
            self.entries
        )
        self._entry_dict = {e.composition.reduced_formula: e for e in self._min_entries}
        self._border_hyperplanes = self._get_border_hyperplanes()
        (
            self._hyperplanes,
            self._hyperplane_entries,
        ) = self._get_hyperplanes_and_entries()

        if self.dim < 2:
            raise ValueError(
                "ChemicalPotentialDiagram currently requires phase "
                "diagrams with 2 or more elements!"
            )

        if len(self.el_refs) != self.dim:
            missing = set(self.elements).difference(self.el_refs.keys())
            raise ValueError(
                f"There are no entries for the terminal elements: {missing}"
            )
        self._hs_int = self._get_halfspace_intersection()

        num_hyperplanes = len(self._hyperplanes)
        num_border_hyperplanes = len(self._border_hyperplanes)

        self._border_hyperplane_indices = list(
            range(num_hyperplanes, num_hyperplanes + num_border_hyperplanes)
        )
        self._metastable_domains: dict[str, list] = {}  # for caching

    def shortest_domain_distance(self, f1: str, f2: str, offset: float = 0.0) -> float:
        """
        Returns the chemical potential distance between two phase domains. Also works
        for metastable phases (see metastable_domains property).

        Args:
            f1: chemical formula of phase 1
            f2: chemical formula of phase 2
            offset: an optional offset (eV/atom) to add to the calculated distance. See
                get_offset() method.

        Returns:
            Shortest distance between domain boundaries in the full (hyper)dimensional
            space, calculated using KDTree.
        """

        if f1 in self.domains:
            pts1 = self.domains[f1]
        else:
            pts1 = self._get_metastable_domain(f1)

        if f2 in self.domains:
            pts2 = self.domains[f2]
        else:
            pts2 = self._get_metastable_domain(f2)

        tree = KDTree(pts1)

        return min(tree.query(pts2)[0]) + offset

    def get_offset(self, entry: PDEntry) -> float:
        """
        For a given entry, returns the distance between its hyperplane and the surface
        of the chemical potential diagram. This allows one to represent the energy above
        hull in chemical potential space. Returns zero for stable entries.

        Args:
            entry: A stable or metastable entry within the chemical potential diagram
        Returns:
            Offset in chemical potential distance (eV/atom)
        """
        if (
            entry in self._min_entries
            and entry.composition.reduced_formula in self.domains
        ):
            offset = 0.0
        else:
            e_above_hull = self._entry_set.get_e_above_hull(entry)
            hyperplane = self._get_hyperplane(entry)
            offset = self._get_distance_between_parallel_hyperplanes(
                hyperplane[:-1], e_above_hull
            )

        return offset

    @cached_property
    def domains(self) -> dict[str, np.ndarray]:
        """
        Mapping of formulas to array of domain boundary points. Cached for quicker
        calculations.
        """
        return self._get_domains()

    @property
    def metastable_domains(self) -> dict[str, np.ndarray]:
        """
        Gets a dictionary of the chemical potential domains for metastable chemical
        formulas. This corresponds to the domains of the relevant phases if they were
        just barely thermodynamically stable (on the hull).
        """
        return {
            e.composition.reduced_formula: self._get_metastable_domain(
                e.composition.reduced_formula
            )
            for e in self._min_entries
            if e.composition.reduced_formula not in self.domains
        }

    @property
    def hs_int(self) -> HalfspaceIntersection:
        """
        Returns the scipy HalfSpaceIntersection object used to calculate all domains.
        """
        return self._hs_int

    def _get_halfspace_intersection(self):
        hs_hyperplanes = np.vstack([self._hyperplanes, self._border_hyperplanes])
        interior_point = np.min(self.lims, axis=1) + 1e-1
        return HalfspaceIntersection(hs_hyperplanes, interior_point)

    def _get_domains(self) -> dict[str, np.ndarray]:
        """Returns a dictionary of chemical potential domains as {formula:
        np.ndarray}"""
        domains: dict[str, list] = {
            entry.composition.reduced_formula: [] for entry in self._hyperplane_entries
        }
        entries = self._hyperplane_entries

        for intersection, facet in zip(
            self.hs_int.intersections, self.hs_int.dual_facets
        ):
            for v in facet:
                if v not in self._border_hyperplane_indices:
                    this_entry = entries[v]
                    formula = this_entry.composition.reduced_formula
                    domains[formula].append(intersection)

        return {k: np.array(v) for k, v in domains.items() if v}

    def _get_hyperplanes_and_entries(self) -> tuple[np.ndarray, list[PDEntry]]:
        """Returns both the array of hyperplanes, as well as a list of the minimum
        entries"""
        data = np.array([self._get_hyperplane(e) for e in self._min_entries])
        vec = [self.el_refs[el].energy_per_atom for el in self.elements] + [1]
        form_e = -np.dot(data, vec)

        inds = np.where(form_e < -PhaseDiagram.formation_energy_tol)[0].tolist()

        inds.extend([self._min_entries.index(el) for el in self.el_refs.values()])

        hyperplanes = data[inds]
        hyperplane_entries = [self._min_entries[i] for i in inds]

        return hyperplanes, hyperplane_entries

    def _get_hyperplane(self, entry):
        data = np.array(
            [entry.composition.get_atomic_fraction(el) for el in self.elements]
            + [-entry.energy_per_atom]
        )
        return data

    def _get_metastable_domain(self, formula, tol=1e-5):
        """Returns the metastable domain for a given formula. Tol is passed to
        GibbsEntrySet.get_stabilized_entry and will affect the size of the domain."""
        if formula in self._metastable_domains:
            return self._metastable_domains[formula]

        orig_entry = self._entry_set.get_min_entry_by_formula(formula)
        new_entry = self._entry_set.get_stabilized_entry(orig_entry, tol=tol)
        self._entry_set.add(new_entry)
        cpd = ChemicalPotentialDiagram(self._entry_set, default_min_limit=-500)

        try:
            metastable_domain = cpd.domains[formula]
        except KeyError as exc:
            # sometimes if the entry is exactly on the hull it fails, so set force=True
            # and make bigger tolerance
            self._entry_set.remove(new_entry)
            new_entry = self._entry_set.get_stabilized_entry(
                orig_entry, tol=1e-2, force=True
            )
            self._entry_set.add(new_entry)
            cpd = ChemicalPotentialDiagram(self._entry_set, default_min_limit=-500)

            try:
                metastable_domain = cpd.domains[formula]
            except KeyError:
                raise ValueError(
                    "Failed even after attempted fix. Metastable domain for"
                    f" {formula} can not be created!"
                ) from exc

        self._metastable_domains[formula] = metastable_domain
        self._entry_set.remove(new_entry)

        return metastable_domain

    @staticmethod
    def _get_distance_between_parallel_hyperplanes(a, delta_b):
        """Returns the distance between two parallel hyperplanes"""
        return np.abs(delta_b) / np.linalg.norm(a)



"""
A reaction class that builds reactions based on ComputedEntry objects under the presence
of an open entry (e.g. O2), and provides information about reaction thermodynamics
computed as changes in grand potential.
"""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.analysis.phase_diagram import GrandPotPDEntry
from pymatgen.core.composition import Element

from rxn_network.core import Composition
from rxn_network.reactions.computed import ComputedReaction

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import ComputedEntry


class OpenComputedReaction(ComputedReaction):
    """
    Extends the ComputedReaction class to add support for "open" reactions,
    where the reaction energy is calculated as a change in grand potential.
    """

    def __init__(
        self,
        entries: list[ComputedEntry],
        coefficients: np.ndarray | list[float],
        chempots: dict[Element, float],
        data: dict | None = None,
        lowest_num_errors: int | float = 0,
    ):
        """
        Args:
            entries: List of ComputedEntry objects.
            coefficients: List of reaction coefficients.
            chempots: Dict of chemical potentials corresponding to open elements.
            data: Optional dict of data.
            lowest_num_errors: number of "errors" encountered during reaction
                balancing
        """
        super().__init__(
            entries=entries,
            coefficients=coefficients,
            data=data,
            lowest_num_errors=lowest_num_errors,
        )

        self.chempots = chempots
        self.open_elems = list(chempots.keys())

        grand_entries = []
        for e in entries:
            comp = e.composition.reduced_composition
            if len(comp.elements) == 1 and comp.elements[0] in self.open_elems:
                grand_entries.append(e)
            else:
                grand_entries.append(GrandPotPDEntry(e, chempots))

        self.grand_entries = grand_entries

    @classmethod
    def balance(  # type: ignore
        cls,
        reactant_entries: list[ComputedEntry],
        product_entries: list[ComputedEntry],
        chempots: dict[Element, float],
        data: dict | None = None,
    ) -> OpenComputedReaction:
        """
        Balances and returns a new ComputedReaction.

        Reactants and products to be specified as a collection (list, set, etc.) of
        ComputedEntry objects.

        A dictionary of open elements and their corresponding chemical potentials must
        be supplied.

        Args:
            reactant_entries: List of reactant entries
            product_entries: List of product entries
            chempots: Dict of chemical potentials corresponding to open
                element(s)
            data: Optional dict of data
        """
        reactant_comps = [e.composition.reduced_composition for e in reactant_entries]
        product_comps = [e.composition.reduced_composition for e in product_entries]
        coefficients, lowest_num_errors, num_constraints = cls._balance_coeffs(
            reactant_comps, product_comps
        )

        if not data:
            data = {}
        data["num_constraints"] = num_constraints

        entries = list(reactant_entries) + list(product_entries)

        kwargs = {
            "entries": entries,
            "coefficients": coefficients,
            "data": data,
            "lowest_num_errors": lowest_num_errors,
        }

        if not chempots:
            rxn = ComputedReaction(**kwargs)  # type: ignore
        else:
            rxn = cls(chempots=chempots, **kwargs)  # type: ignore

        return rxn

    def get_new_temperature(self, new_temperature: float) -> OpenComputedReaction:
        """
        Returns a new reaction with the temperature changed.

        Args:
            new_temperature: New temperature in Kelvin
        """
        try:
            new_entries = [e.get_new_temperature(new_temperature) for e in self.entries]
        except AttributeError as e:
            raise AttributeError(
                "One or more of the entries in the reaction is not associated with a"
                " temperature. Please use the GibbsComputedEntry class for all entries"
                " in the reaction."
            ) from e

        return OpenComputedReaction(
            new_entries,
            self.coefficients,
            chempots=self.chempots,
            data=self.data,
            lowest_num_errors=self.lowest_num_errors,
        )

    @property
    def energy(self) -> float:
        """
        Returns (float):
            The calculated reaction energy.
        """
        calc_energies: dict[Composition, float] = {}

        for entry in self.grand_entries:
            attr = "composition"
            if isinstance(entry, GrandPotPDEntry):
                attr = "original_comp"

            comp, factor = getattr(entry, attr).get_reduced_composition_and_factor()
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), entry.energy / factor
            )

        return sum(
            amt * calc_energies[c]
            for amt, c in zip(self.coefficients, self.compositions)
        )

    @property
    def elements(self) -> list[Element]:
        """
        List of elements in the reaction
        """
        return list(
            set(el for comp in self.compositions for el in comp.elements)
            - set(self.open_elems)
        )

    @property
    def total_chemical_system(self) -> str:
        """
        Chemical system string, including open elements
        """
        return "-".join(
            sorted([str(e) for e in set(self.elements) | set(self.open_elems)])
        )

    def copy(self) -> OpenComputedReaction:
        """
        Returns a copy of the OpenComputedReaction object.
        """
        return OpenComputedReaction(
            self.entries,
            self.coefficients,
            self.chempots,
            self.data,
            self.lowest_num_errors,
        )

    def reverse(self):
        """
        Returns a copy of reaction with reactants/products swapped
        """
        return OpenComputedReaction(
            self.entries,
            -1 * self.coefficients,
            self.chempots,
            self.data,
            self.lowest_num_errors,
        )

    @cached_property
    def reactant_atomic_fractions(self) -> dict:
        """
        Returns the atomic mixing ratio of reactants in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        return {
            c.reduced_composition: -coeff
            * sum(c[el] for el in self.elements)
            / self.num_atoms
            for c, coeff in self.reactant_coeffs.items()
        }

    @cached_property
    def product_atomic_fractions(self) -> dict:
        """
        Returns the atomic mixing ratio of reactants in the reaction
        """
        if not self.balanced:
            raise ValueError("Reaction is not balanced")

        return {
            c.reduced_composition: sum(c[el] for el in self.elements) / self.num_atoms
            for c, coeff in self.product_coeffs.items()
        }

    @classmethod
    def from_computed_rxn(
        cls, reaction: ComputedReaction, chempots: dict[Element, float]
    ) -> OpenComputedReaction:
        return cls(
            entries=reaction.entries.copy(),
            coefficients=reaction.coefficients.copy(),
            chempots=chempots,
            data=reaction.data.copy(),
            lowest_num_errors=reaction.lowest_num_errors,
        )

    def as_dict(self) -> dict:
        """
        Returns a dictionary representation of the reaction.
        """
        d = super().as_dict()
        d["chempots"] = {el.symbol: u for el, u in self.chempots.items()}
        return d

    @classmethod
    def from_dict(cls, d) -> OpenComputedReaction:
        """
        Returns an OpenComputedReaction object from a dictionary representation.
        """
        d["chempots"] = {Element(symbol): u for symbol, u in d["chempots"].items()}
        return super().from_dict(d)

    def __repr__(self) -> str:
        cp = f"({','.join([f'mu_{e}={m}' for e, m in self.chempots.items()])})"
        return f"{super().__repr__()} {cp}"



"""
This module contains functions for plotting experimental reaction pathway data.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import pandas
from monty.json import MSONable
from pymatgen.analysis.phase_diagram import PhaseDiagram
from scipy.ndimage import median_filter
from tqdm import tqdm

from rxn_network.core import Composition
from rxn_network.entries.entry_set import GibbsEntrySet

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element
    from pymatgen.entries.computed_entries import ComputedEntry


class PathwayPlotter(MSONable):
    """
    WARNING: This is an EXPERIMENTAL CLASS. Use at your own risk.

    Helper class for plotting a reaction pathway and the corresponding energy cascade.
    """

    def __init__(
        self,
        phase_amounts: dict[str, list[float]],
        temps: list[float],
        apply_smoothing: bool = True,
    ):
        """
        Args:
            phase_amounts: Dicts with format {phase: [amounts]}
            temps: list of temperatures
            apply_smoothing: Whether to smooth the data. Default is True.
        """
        self._phase_amounts = phase_amounts
        self._temps = temps
        self._apply_smoothing = apply_smoothing

        self._formulas = list(phase_amounts.keys())
        self._df = self._get_phase_df()

        self._num_atoms_df = self._get_num_atoms_df()

    def plot_pathway(self):
        """
        Returns a plot of the pathway by calling DataFrame.plot().
        """
        return self.df.plot()

    def plot_energy_cascade(self, entries: list[ComputedEntry] | GibbsEntrySet):
        """
        Returns a plot of the energy cascade given a list of entries.

        Args:
            entries: List of entries or GibbsEntrySet.

        """
        energies = self._get_energies(entries)
        energies_df = pandas.DataFrame(energies).T
        ground_state_energies = energies_df.pop("ground_state")

        gibbs_arr = (
            self.df.values
            * energies_df.values
            / self.num_atoms_df.sum(axis=1).values.reshape(-1, 1)
        )

        g_df = pandas.DataFrame(gibbs_arr, columns=self.df.columns, index=self.df.index)

        total_g = g_df.sum(axis=1)
        total_g = total_g - ground_state_energies

        plot = total_g.plot(backend="matplotlib", style="o")
        plt.xlabel("Temperature (K)", {"size": 11})
        plt.ylabel(r"Gibbs Free Energy, $G$ (eV/atom)", {"size": 11})

        return plot

    def _get_energies(self, entries):
        """
        Interal method: returns a list of energies for each phase
        """
        all_energies = {}
        formulas = self.df.columns.to_list()

        compositions = self.compositions

        for temp in tqdm(self.df.index):
            all_energies[temp] = {}

            gibbs_entries = GibbsEntrySet.from_computed_entries(entries, temp)
            pd = PhaseDiagram(gibbs_entries)

            ground_state_energy = pd.get_hull_energy(compositions[temp])
            all_energies[temp]["ground_state"] = ground_state_energy

            for f in formulas:
                comp = Composition(f).reduced_composition
                energy = (
                    pd.get_form_energy_per_atom(
                        gibbs_entries.get_min_entry_by_formula(comp.formula)
                    )
                    * comp.num_atoms
                )
                all_energies[temp][f] = energy

        return all_energies

    def _get_phase_df(self):
        """
        Returns a dataframe of phase amounts
        """
        phase_df = pandas.DataFrame(self._phase_amounts, index=self._temps)
        if self._apply_smoothing:
            phase_df = phase_df.apply(median_filter, axis=0, size=3)

        return phase_df

    def _get_num_atoms_df(self):
        """
        Returns a dataframe of the number of atoms in each phase
        """
        el_dict = {str(e): [] for e in self.elems}

        for idx, f in enumerate(self.df.columns):
            comp = Composition(f)
            col = self.df.iloc[:, idx]

            for el in el_dict.keys():
                el_dict[el].append((col * comp.get_el_amt_dict()[el]).fillna(0))

        for el in el_dict:
            el_dict[el] = sum(el_dict[el])

        el_df = pandas.DataFrame(el_dict)

        return el_df

    @property
    def elems(self) -> list[Element]:
        """
        Returns a list of elements in the pathway
        """
        return list({e for f in self.formulas for e in Composition(f).elements})

    @property
    def num_atoms_df(self) -> pandas.DataFrame:
        """
        Returns a dataframe of the number of atoms in each phase
        """
        return self._num_atoms_df

    @property
    def formulas(self) -> list[str]:
        """
        Returns a list of formulas in the pathway
        """
        return self._formulas

    @property
    def df(self) -> pandas.DataFrame:
        """
        Returns a dataframe of the pathway
        """
        return self._df

    @property
    def compositions(self) -> pandas.Series:
        """
        Returns the composition of the pathway
        """
        comps = [
            Composition(i).fractional_composition
            for i in self.num_atoms_df.to_dict(  # pylint: disable=not-an-iterable
                "records"
            )
        ]
        return pandas.Series(comps, index=self.num_atoms_df.index)



"""
Implements a class for conveniently and efficiently storing sets of ComputedReaction
objects which share entries.
"""
from __future__ import annotations

from collections import OrderedDict
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Collection, Iterable

import numpy as np
import ray
from monty.json import MSONable
from pandas import DataFrame
from pymatgen.core.composition import Element
from tqdm import tqdm

from rxn_network.core import Composition
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.utils.funcs import get_logger, grouper
from rxn_network.utils.ray import initialize_ray, to_iterator

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import ComputedEntry

    from rxn_network.costs.base import CostFunction

logger = get_logger(__name__)


class ReactionSet(MSONable):
    """
    A lightweight class for storing large sets of ComputedReaction objects.
    Automatically represents a set of reactions as an array of coefficients with
    a second array linking to a corresponding list of shared entries. This is useful for
    dumping large amounts of reaction data to a database.

    Note: this is not a true "set"; there is the option for filtering duplicates but it
        is not explicitly required.
    """

    def __init__(
        self,
        entries: list[ComputedEntry],
        indices: dict[int, np.ndarray],
        coeffs: dict[int, np.ndarray],
        open_elem: str | Element | None = None,
        chempot: float = 0.0,
        all_data: dict[int, np.ndarray] | None = None,
    ):
        """
        Args:
            entries: List of ComputedEntry objects shared by reactions
            indices: Array indexing the entry list; gets entries used by each
                reaction object
            coeffs: Array of all reaction coefficients
            open_elem: Open element, e.g., "O"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
            all_data: Optional list of data for each reaction
        """
        self.entries = entries
        self.indices = indices
        self.coeffs = coeffs
        self.open_elem = open_elem
        self.chempot = chempot
        if all_data is None:
            self.all_data = {int(i): np.array([]) for i in self.indices}
        else:
            self.all_data = all_data

        if not all(isinstance(k, int) for k in self.indices.keys()) or not all(
            isinstance(v, np.ndarray) for v in self.indices.values()
        ):
            self.indices = {
                int(size): np.array(arr) for size, arr in self.indices.items()
            }

        if not all(isinstance(k, int) for k in self.coeffs.keys()) or not all(
            isinstance(v, np.ndarray) for v in self.coeffs.values()
        ):
            self.coeffs = {
                int(size): np.array(arr) for size, arr in self.coeffs.items()
            }

        if not all(isinstance(k, int) for k in self.all_data.keys()) or not all(
            isinstance(v, np.ndarray) for v in self.all_data.values()
        ):
            self.all_data = {
                int(size): np.array(arr) for size, arr in self.all_data.items()
            }

        self.open_elem = open_elem
        self.chempot = chempot

        self.mu_dict = None

        if open_elem:
            self.mu_dict = {Element(open_elem): chempot}  # type: ignore

    def get_rxns(
        self,
    ) -> Iterable[ComputedReaction | OpenComputedReaction]:
        """
        Generator for all ComputedReaction objects or OpenComputedReaction objects (when
        open element and chempot are specified) for the reaction set.
        """
        return self._get_rxns_by_indices(
            idxs={i: slice(0, len(c)) for i, c in self.indices.items()}
        )

    @classmethod
    def from_rxns(
        cls,
        rxns: Collection[ComputedReaction | OpenComputedReaction],
        entries: Collection[ComputedEntry] | None = None,
        open_elem: str | Element | None = None,
        chempot: float = 0.0,
        filter_duplicates: bool = False,
    ) -> ReactionSet:
        """
        Initiate a ReactionSet object from a list of reactions. Including a list of
        unique entries saves some computation time.

        Args:
            rxns: List of ComputedReaction-like objects.
            entries: Optional list of ComputedEntry objects
            open_elem: Open element, e.g. "O2"
            chempot: Chemical potential (mu) of open element in equation: Phi = G - mu*N
        """

        if not entries:
            entries = cls._get_unique_entries(rxns)

        entries = sorted(list(set(entries)), key=lambda r: r.composition)

        # need to index by unique ID in case mixing different temperatures
        all_entry_indices = {}
        for idx, entry in enumerate(entries):
            if hasattr(entry, "unique_id"):
                all_entry_indices[entry.unique_id] = idx
            else:
                all_entry_indices[entry.entry_id] = idx

        # group by reaction size to ensure rectangular matrices
        indices, coeffs, data = {}, {}, {}  # type: ignore

        for rxn in rxns:
            size = len(rxn.entries)

            rxn_indices = []
            for e in rxn.entries:
                if hasattr(e, "unique_id"):
                    rxn_indices.append(all_entry_indices[e.unique_id])
                else:
                    rxn_indices.append(all_entry_indices[e.entry_id])

            if size not in indices:
                indices[size] = []
                coeffs[size] = []
                data[size] = []

            indices[size].append(rxn_indices)
            coeffs[size].append(rxn.coefficients)
            data[size].append(rxn.data)

        for size in indices:
            indices[size] = np.array(indices[size])
            coeffs[size] = np.array(coeffs[size])
            data[size] = np.array(data[size])

        all_open_elems: set[Element] = set()
        all_chempots: set[float] = set()

        if (
            all(r.__class__.__name__ == "OpenComputedReaction" for r in rxns)
            and not open_elem
        ):
            for r in rxns:
                all_open_elems.update(r.chempots.keys())
                all_chempots.update(r.chempots.values())

            if len(all_chempots) == 1 and len(all_open_elems) == 1:
                chempot = all_chempots.pop()
                open_elem = all_open_elems.pop()

        rxn_set = cls(
            entries=entries,
            indices=indices,
            coeffs=coeffs,
            open_elem=open_elem,
            chempot=chempot,
            all_data=data,
        )

        if filter_duplicates:
            rxn_set = rxn_set.filter_duplicates()

        return rxn_set

    @lru_cache(maxsize=1)
    def to_dataframe(
        self,
        cost_function: CostFunction,
        target: Composition | None = None,
        calculate_uncertainties: bool = False,
        calculate_separable: bool = False,
    ) -> DataFrame:
        """
        Make a dataframe of reactions from a ReactionSet object.

        Args:
            cost_function: Cost function to use for evaluating reaction costs
            target: Optional target composition (used to determine added elements)
            calculate_uncertainties: Whether to calculate uncertainties (dE column)
            calculate_separable: Whether to calculate if the reaction products are
                separable (see ComputedReaction.is_separable)

        Returns:
            Pandas DataFrame with columns:
                rxn: Reaction object
                energy: reaction energy in eV/atom
                dE (optional): uncertainty in reaction energy in eV/atom
                added_elems (optional): List of added elements
                separable (optional): whether reaction products are separable
                cost: Cost of reaction
                other: any other data associated with reaction

        """
        data: dict[str, Any] = OrderedDict({k: [] for k in ["rxn", "energy"]})
        attrs = []

        calculate_e_above_hulls = False
        determine_theoretical = False

        for r in self.get_rxns():
            attrs = list(r.data.keys())  # get extra attributes from first reaction
            entry_data = r.entries[0].data
            if entry_data.get("e_above_hull") is not None:
                calculate_e_above_hulls = True
                data["max_e_hull_reactants"] = []
                data["max_e_hull_products"] = []
            if "icsd_ids" in entry_data or "theoretical" in entry_data:
                determine_theoretical = True
                data["num_theoretical_reactants"] = []
                data["num_theoretical_products"] = []

            break

        target = Composition(target) if target else None

        if "num_constraints" in attrs:
            attrs.remove("num_constraints")
        if calculate_uncertainties:
            data["dE"] = []
        if target:
            data["added_elems"] = []
            if calculate_separable:
                data["separable"] = []

        data["max_num_precursor_elems"] = []

        data.update({k: [] for k in attrs + ["cost"]})

        for rxn in self.get_rxns():
            data["rxn"].append(rxn)
            data["energy"].append(rxn.energy_per_atom)
            if calculate_uncertainties:
                data["dE"].append(rxn.energy_uncertainty_per_atom)
            if target:
                data["added_elems"].append(self._get_added_elems(rxn, target))
                if calculate_separable:
                    data["separable"].append(rxn.is_separable(target))

            if calculate_e_above_hulls:
                data["max_e_hull_reactants"].append(
                    max(
                        e.data.get("e_above_hull", 0.0) or 0.0
                        for e in rxn.reactant_entries
                    )
                )
                data["max_e_hull_products"].append(
                    max(
                        e.data.get("e_above_hull", 0.0) or 0.0
                        for e in rxn.product_entries
                    )
                )
            if determine_theoretical:
                data["num_theoretical_reactants"].append(
                    sum(bool(not e.is_experimental) for e in rxn.reactant_entries)
                )
                data["num_theoretical_products"].append(
                    sum(bool(not e.is_experimental) for e in rxn.product_entries)
                )

            data["max_num_precursor_elems"].append(
                max(len(precursor.elements) for precursor in rxn.reactants)
            )

            for attr in attrs:
                data[attr].append(rxn.data.get(attr))

            data["cost"].append(cost_function.evaluate(rxn))

        df = DataFrame(data).sort_values("cost").reset_index(drop=True)
        return df

    def calculate_costs(
        self,
        cf: CostFunction,
    ) -> list[float]:
        """
        Evaluate a cost function on an acquired set of reactions.

        Args:
            cf: CostFunction object, e.g. Softplus()
        """
        return [cf.evaluate(rxn) for rxn in self.get_rxns()]

    def add_rxns(self, rxns: Collection[ComputedReaction | OpenComputedReaction]):
        """
        Return a new ReactionSet with the reactions added.

        Warning: all new reactions must only have entires contained in the entries of
        the current reaction set.
        """
        new_rxn_set = ReactionSet.from_rxns(rxns, entries=self.entries)

        return self.add_rxn_set(new_rxn_set)

    def add_rxn_set(self, rxn_set: ReactionSet) -> ReactionSet:
        """Adds a new reaction set to current reaction set.

        Warning: new reaction set must have the same entries as the current reaction
        set.

        """
        if self.entries != rxn_set.entries:
            raise ValueError(
                "Reaction sets must have identical entries property to combine."
            )
        open_elem = self.open_elem
        chempot = self.chempot

        new_indices = {}
        new_coeffs = {}
        new_all_data = {}

        for size in set(list(self.indices.keys()) + list(rxn_set.indices.keys())):
            if size in self.indices and size in rxn_set.indices:
                new_indices[size] = np.concatenate(
                    (self.indices[size], rxn_set.indices[size])
                )
                new_coeffs[size] = np.concatenate(
                    (self.coeffs[size], rxn_set.coeffs[size])
                )
                new_all_data[size] = np.concatenate(
                    (self.all_data[size], rxn_set.all_data[size])
                )
            elif size in self.indices:
                new_indices[size] = self.indices[size]
                new_coeffs[size] = self.coeffs[size]
                new_all_data[size] = self.all_data[size]
            elif size in rxn_set.indices:
                new_indices[size] = rxn_set.indices[size]
                new_coeffs[size] = rxn_set.coeffs[size]
                new_all_data[size] = rxn_set.all_data[size]

        return ReactionSet(
            self.entries, new_indices, new_coeffs, open_elem, chempot, new_all_data
        )

    def get_rxns_by_reactants(
        self, reactants: list[str], return_set: bool = False
    ) -> Iterable[ComputedReaction | OpenComputedReaction]:
        """
        Return a list of reactions with the given reactants.
        """
        reactants = [Composition(r).reduced_formula for r in reactants]

        reactant_indices = list(
            {
                idx
                for idx, e in enumerate(self.entries)
                if e.composition.reduced_formula in reactants
            }
        )

        if not reactant_indices:
            return []

        idxs = {}  # type: ignore
        for size, indices in self.indices.items():
            idxs[size] = []
            contains_reactants = np.isin(indices, reactant_indices).any(axis=1)
            for idx, coeffs, indices in zip(
                np.argwhere(contains_reactants).flatten(),
                self.coeffs[size][contains_reactants],
                self.indices[size][contains_reactants],
            ):
                r_indices = {i for c, i in zip(coeffs, indices) if c < 1e-12}
                if r_indices.issubset(reactant_indices):
                    idxs[size].append(idx)

        if return_set:
            return self._get_rxn_set_by_indices(idxs)

        return self._get_rxns_by_indices(idxs)

    def get_rxns_by_product(self, product: str, return_set: bool = False):
        """
        Return a list of reactions which contain the given product formula.

        Args:
            product: The product's formula
            return_set: Whether to return the identified reactions in the form of a
                ReactionSet object. Defaults to False.

        """
        product = Composition(product).reduced_formula

        product_index = None
        for idx, e in enumerate(self.entries):
            if e.composition.reduced_formula == product:
                product_index = idx
                break

        if not product_index:
            return []

        idxs = {}  # type: ignore
        for size, indices in self.indices.items():
            idxs[size] = []
            contains_product = np.isin(indices, product_index).any(axis=1)

            for idx, coeffs, indices in zip(
                np.argwhere(contains_product).flatten(),
                self.coeffs[size][contains_product],
                self.indices[size][contains_product],
            ):
                p_indices = {i for c, i in zip(coeffs, indices) if c > -1e-12}
                if product_index in p_indices:
                    idxs[size].append(idx)

        if return_set:
            return self._get_rxn_set_by_indices(idxs)

        return self._get_rxns_by_indices(idxs)

    def filter_duplicates(
        self,
        ensure_rxns: list[ComputedReaction | OpenComputedReaction] | None = None,
        parallelize: bool = True,
    ) -> ReactionSet:
        """
        Returns a new ReactionSet object with duplicate reactions removed.

        NOTE: Duplicate reactions include those that are multiples of each other. For
        example, if a reaction set contains both A + B -> C and 2A + 2B -> 2C, the
        second reaction will be removed.

        Args:
            ensure_rxns: An optional list of reactions to ensure are contained within
                the filtered set that is returned. This is important for some cases
                (e.g., pathfinding), where you expect a certain reaction object to be in
                the set.
            parallelize: Whether to parallelize duplicate checking with Ray. This can be
                a slow procedure otherwise. Defaults to True.
        """
        if parallelize:
            if not ray.is_initialized():
                initialize_ray()
            num_cpus = int(ray.cluster_resources()["CPU"]) - 1
            chunk_refs = []

        ensure_idxs = {size: [] for size in self.indices}  # type: ignore
        idxs_to_keep = {size: [] for size in self.indices}  # type: ignore

        # identify indices to keep
        for rxn in ensure_rxns or []:
            rxn_idxs = np.array([self.entries.index(e) for e in rxn.entries])
            rxn_coeffs = rxn.coefficients

            size = len(rxn_idxs)

            # filter by first column to save some time (full filter is long)
            quick_filter = np.argwhere(
                rxn_idxs[0] == self.indices[size][:, 0]
            ).flatten()

            for i, idxs, coeffs in zip(
                quick_filter,
                self.indices[size][quick_filter],
                self.coeffs[size][quick_filter],
            ):
                if (idxs == rxn_idxs).all() and (coeffs == rxn_coeffs).all():
                    ensure_idxs[size].append(i)
                    break
            else:
                raise ValueError(f"{rxn} not found in ReactionSet!")

        for size in sorted(self.indices.keys()):
            ensure_idxs_to_keep = ensure_idxs[size]

            # reordering columns in ascending entry idx
            column_sorting_indices = np.argsort(self.indices[size], axis=1)
            indices = np.take_along_axis(
                self.indices[size], column_sorting_indices, axis=1
            )
            coeffs = np.take_along_axis(
                self.coeffs[size], column_sorting_indices, axis=1
            )

            _, inverse_indices = np.unique(
                indices,
                axis=0,
                return_inverse=True,
            )

            keep = []
            group_indices = []

            for i in np.split(
                np.argsort(inverse_indices),
                np.cumsum(np.unique(inverse_indices, return_counts=True)[1])[:-1],
            ):  # this was suggested by ChatGPT
                if len(i) == 1:
                    keep.append(i[0])
                else:
                    group_indices.append(i)

            num_groups = len(group_indices)
            idxs_to_keep[size] = keep

            if num_groups == 0:
                continue

            if parallelize:
                chunk_size = (num_groups // num_cpus) + 1

                coeffs = ray.put(coeffs)

                for chunk in grouper(
                    group_indices,
                    chunk_size,
                    fillvalue=None,
                ):
                    chunk_refs.append(
                        _process_duplicates_ray.remote(
                            size, chunk, coeffs, ensure_idxs_to_keep
                        )
                    )
            else:
                idxs_to_keep[size].extend(
                    _process_duplicates(group_indices, coeffs, ensure_idxs_to_keep)
                )

        if parallelize:
            for result in tqdm(
                to_iterator(chunk_refs),
                total=len(chunk_refs),
                desc="Filtering duplicates",
            ):
                size, data = result
                idxs_to_keep[size].extend(data)

        return self._get_rxn_set_by_indices(idxs_to_keep)

    def set_chempot(self, open_el: str | Element | None, chempot: float) -> ReactionSet:
        """
        Returns a new ReactionSet containing the same reactions as this ReactionSet but
        with a grand potential change recalculated under the constraint defined by the
        provided open element and its chemical potential.

        Args:
            open_el: The element to be considered open.
            chempot: The open element's chemical potential (for use in energy change
                calculation)

        Returns:
            ReactionSet: A new ReactionSet containing reactions with the recalculated
                energies.
        """
        return ReactionSet(
            self.entries, self.indices, self.coeffs, open_el, chempot, self.all_data
        )

    def set_new_temperature(self, new_temp: float) -> ReactionSet:
        """
        Returns a new ReactionSet containing the same reactions as this ReactionSet but
        with a recalculated Gibb's/Grand potential change reflecting formation energies
        calculated at the provided temperature.

        Args:
            new_temp: The temperature for which new reaction energies should be
                calculated.

        Returns:
            The new ReactionSet containing the recalculated reactions.
        """
        new_rxns = []
        for rxn in self.get_rxns():
            new_rxns.append(rxn.get_new_temperature(new_temp))

        return ReactionSet.from_rxns(
            new_rxns, open_elem=self.open_elem, chempot=self.chempot
        )

    def _get_rxns_by_indices(
        self, idxs
    ) -> Iterable[ComputedReaction | OpenComputedReaction]:
        """
        Return a list of reactions with the given indices.
        """

        for size, idx_arr in idxs.items():
            if not idx_arr:
                idx_arr = slice(0, 0)
            elif not isinstance(idx_arr, np.ndarray) and not isinstance(idx_arr, slice):
                idx_arr = np.array(idx_arr)

            for indices, coeffs, data in zip(
                self.indices[size][idx_arr],
                self.coeffs[size][idx_arr],
                self.all_data[size][idx_arr],
            ):
                entries = [self.entries[i] for i in indices]
                if self.mu_dict:
                    rxn = OpenComputedReaction(
                        entries=entries,
                        coefficients=coeffs,
                        data=data,
                        chempots=self.mu_dict,
                    )
                else:
                    rxn = ComputedReaction(
                        entries=entries, coefficients=coeffs, data=data
                    )

                yield rxn

    def _get_rxn_set_by_indices(self, idxs) -> ReactionSet:
        """
        Return a list of reactions with the given indices. This is the backbone of other
        reaction subset generation methods.
        """
        return ReactionSet(
            entries=self.entries,
            indices={
                size: self.indices[size][idx_arr] for size, idx_arr in idxs.items()
            },
            coeffs={size: self.coeffs[size][idx_arr] for size, idx_arr in idxs.items()},
            open_elem=self.open_elem,
            chempot=self.chempot,
            all_data={
                size: self.all_data[size][idx_arr] for size, idx_arr in idxs.items()
            },
        )

    @staticmethod
    def _get_added_elems(
        rxn: ComputedReaction | OpenComputedReaction, target: Composition | str
    ) -> str:
        """
        Get list of added elements for a reaction.

        Args:
            rxn: Reaction object
            target: target composition (used to determine added elements)

        """
        target = Composition(target)
        chemsys_prop = (
            "total_chemical_system"
            if rxn.__class__.__name__ == "OpenComputedReaction"
            else "chemical_system"
        )

        added_elems = set(getattr(rxn, chemsys_prop).split("-")) - set(
            target.chemical_system.split("-")
        )
        added_elems_str = "-".join(sorted(list(added_elems)))

        return added_elems_str

    @staticmethod
    def _get_entry_key(entry: ComputedEntry) -> str:
        """
        Get a unique key for an entry. Assumes that a formula and energy alone should be
        able to define a unique entry.

        Args:
            entry: Entry object

        """
        return f"{entry.composition.reduced_formula}_{round(entry.energy_per_atom, 4)}"

    @staticmethod
    def _get_unique_entries(rxns: Collection[ComputedReaction]) -> set[ComputedEntry]:
        """
        Return only unique entries from reactions
        """
        entries = set()
        for r in rxns:
            entries.update(r.entries)
        return entries

    def __iter__(self):
        """
        Iterate over the reactions in the set.
        """
        return iter(self.get_rxns())

    def __len__(self):
        """
        Return length of reactions stored in the set.
        """
        return sum(len(i) for i in self.indices.values())


def _get_idxs_to_keep(rows, ensure_idxs=None):
    """Looks for reactions with coeffs that are positive multiples of each other."""
    sorted_indices = np.argsort(np.abs(rows).sum(axis=1))  # sort by abs sum
    sorted_rows = rows[sorted_indices]
    if ensure_idxs is not None:
        ensure_idxs = set(ensure_idxs)

    to_keep, to_remove = [], []
    for idx, row in zip(sorted_indices, sorted_rows):
        if idx in to_remove:
            continue

        multiples = sorted_rows / row
        duplicates = np.argwhere(
            np.apply_along_axis(
                lambda r: np.isclose(abs(r[0]), r), axis=1, arr=multiples
            ).all(axis=1)
        ).flatten()

        group = sorted_indices[duplicates]
        if ensure_idxs:
            group_set = set(group)
            overlap = ensure_idxs & group_set

            if overlap:
                if not overlap.issubset(to_keep):
                    difference = group_set - ensure_idxs
                    to_keep.extend(list(overlap))
                    to_remove.extend(list(difference))
            else:
                to_keep.append(group[0])
                to_remove.extend(group[1:])
        else:
            to_keep.append(group[0])
            to_remove.extend(group[1:])

    return sorted(to_keep)


@ray.remote
def _process_duplicates_ray(
    size: int,
    groups: list[np.ndarray],
    coeffs: np.ndarray,
    ensure_idxs: list[int],
) -> tuple[int, list[int]]:
    """
    Process a chunk of reactions to find duplicates. This is a remote function within
    Ray.

    Args:
        size: size of reactions to process
        groups: chunks of reactions to process
        coeffs: corresponding coefficients
        ensure_idxs: indices of reactions to ensure are kept

    Returns:
        List of indices to keep
    """
    idxs_to_keep = _process_duplicates(groups, coeffs, ensure_idxs)
    return size, idxs_to_keep


def _process_duplicates(
    groups: list[np.ndarray],
    coeffs: np.ndarray,
    ensure_idxs: list[int],
) -> list[int]:
    """
    Process a chunk of reactions to find duplicates.

    Args:
        chunk: chunk of reactions to process
        coeffs: corresponding coefficients
        ensure_idxs: indices of reactions to ensure are kept
        size: size of reactions to process

    Returns:
        List of indices to keep
    """
    idxs_to_keep = []
    for group in groups:
        if group is None:
            continue
        possible_duplicate_coeffs = coeffs[group]
        relative_ensure_idxs = []
        for idx in ensure_idxs:
            if idx in group:
                relative_ensure_idxs.append(np.argwhere(group == idx).flatten()[0])
        keep_idxs = _get_idxs_to_keep(possible_duplicate_coeffs, relative_ensure_idxs)
        idxs_to_keep.extend(group[keep_idxs])

    return idxs_to_keep



"""
Implements a reaction pathway solver class which efficiently solves mass balance
equations using matrix operations.
"""
from __future__ import annotations

from abc import ABCMeta
from copy import deepcopy
from itertools import combinations
from typing import TYPE_CHECKING

import numpy as np
import ray
from monty.json import MSONable
from numba import jit
from pymatgen.core.composition import Element
from tqdm import tqdm

from rxn_network.core import Composition
from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)
from rxn_network.pathways.balanced import BalancedPathway
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger, grouper
from rxn_network.utils.ray import initialize_ray, to_iterator

if TYPE_CHECKING:
    from rxn_network.costs.base import CostFunction
    from rxn_network.pathways.base import Pathway
    from rxn_network.reactions.base import Reaction

logger = get_logger(__name__)


class Solver(MSONable, metaclass=ABCMeta):
    """
    Base definition for a pathway solver class.
    """

    def __init__(self, pathways: PathwaySet):
        self._pathways = pathways

        rxns = []
        costs = []

        for path in self._pathways.paths:
            for rxn, cost in zip(path.reactions, path.costs):
                if rxn not in rxns:
                    rxns.append(rxn)
                    costs.append(cost)

        self._reactions = rxns
        self._costs = costs

    @property
    def pathways(self) -> list[Pathway]:
        """Pathways used in solver class"""
        return self._pathways

    @property
    def reactions(self) -> list[Reaction]:
        """Reactions used in solver class"""
        return self._reactions

    @property
    def costs(self) -> list[float]:
        """Costs used in solver class"""
        return self._costs

    @property
    def num_rxns(self) -> int:
        """Length of the reaction list"""
        return len(self.reactions)

    @property
    def num_entries(self) -> int:
        """Length of entry list"""
        return len(self._entries)


class PathwaySolver(Solver):
    """
    Solver that implements an efficient method (using numba) for finding balanced
    reaction pathways from a list of graph-derived reaction pathways (i.e. a list of
    lists of reactions).

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.
    """

    def __init__(
        self,
        pathways: PathwaySet,
        entries: GibbsEntrySet,
        cost_function: CostFunction,
        open_elem: str | Element | None = None,
        chempot: float = 0.0,
        chunk_size: int = 100000,
        batch_size: int | None = None,
    ):
        """
        Args:
            pathways: List of reaction pathways derived from the network.
            entries: GibbsEntrySet containing all entries in the network.
            cost_function: CostFunction object to use for the solver.
            open_elem: Optional element to use for pathways with an open element.
            chempot: Chemical potential to use for pathways with an open element.
                Defaults to 0.0.
            chunk_size: The number of pathways per chunk to use for balancing. Defaults
                to 100,000.
            batch_size: Number of chunks to submit to each CPU at a time. Automatically
                calculated if not set.


        """
        super().__init__(pathways=deepcopy(pathways))
        self._entries = entries
        self.cost_function = cost_function
        self.open_elem = Element(open_elem) if open_elem else None
        self.chempot = chempot
        self.chunk_size = chunk_size
        self.batch_size = batch_size

    def solve(
        self,
        net_rxn: ComputedReaction | OpenComputedReaction,
        max_num_combos: int = 4,
        find_intermediate_rxns: bool = True,
        intermediate_rxn_energy_cutoff: float = 0.0,
        use_basic_enumerator: bool = True,
        use_minimize_enumerator: bool = False,
        filter_interdependent: bool = True,
    ) -> PathwaySet:
        """

        Args:
            net_rxn: The reaction representing the total reaction from precursors to
                final targets.
            max_num_combos: The maximum allowable size of the balanced reaction pathway.
                At values <=5, the solver will start to take a significant amount of
                time to run.
            find_intermediate_rxns: Whether to find intermediate reactions; crucial for
                finding pathways where intermediates react together, as these reactions
                may not occur in the graph-derived pathways. Defaults to True.
            intermediate_rxn_energy_cutoff: An energy cutoff by which to filter down
                intermediate reactions. This can be useful when there are a large number
                of possible intermediates. < 0 means allow only exergonic reactions.
            use_basic_enumerator: Whether to use the BasicEnumerator to find
                intermediate reactions. Defaults to True.
            use_minimize_enumerator: Whether to use the MinimizeGibbsEnumerator to find
                intermediate reactions. Defaults to False.
            filter_interdependent: Whether or not to filter out pathways where reaction
                steps are interdependent. Defaults to True.

        Returns:
            A list of BalancedPathway objects.
        """

        if not net_rxn.balanced:
            raise ValueError(
                "Net reaction must be balanceable to find all reaction pathways."
            )

        if not ray.is_initialized():
            initialize_ray()

        entries_copy = deepcopy(self.entries)
        entries = entries_copy.entries_list
        num_entries = len(entries)

        reactions = deepcopy(self.reactions)
        costs = deepcopy(self.costs)

        precursors = deepcopy(net_rxn.reactant_entries)
        targets = deepcopy(net_rxn.product_entries)

        logger.info(f"Net reaction: {net_rxn} \n")

        if find_intermediate_rxns:
            logger.info("Identifying reactions between intermediates...")

            intermediate_rxns = self._find_intermediate_rxns(
                targets,
                intermediate_rxn_energy_cutoff,
                use_basic_enumerator,
                use_minimize_enumerator,
            )
            intermediate_costs = [
                self.cost_function.evaluate(r) for r in intermediate_rxns.get_rxns()
            ]
            for r, c in zip(intermediate_rxns, intermediate_costs):
                if r not in reactions:
                    reactions.append(r)
                    costs.append(c)

        clean_r_set = ReactionSet.from_rxns(reactions, filter_duplicates=True)
        cleaned_reactions, cleaned_costs = zip(
            *[
                (r, c)
                for r, c in zip(reactions, costs)
                if r in clean_r_set and r != net_rxn
            ]
        )

        net_rxn_vector = net_rxn.get_entry_idx_vector(num_entries)

        num_rxns = len(cleaned_reactions)
        num_cpus = ray.cluster_resources()["CPU"]
        batch_size = self.batch_size or num_cpus - 1

        net_coeff_filter = np.argwhere(net_rxn_vector != 0).flatten()
        net_coeff_filter = ray.put(net_coeff_filter)
        cleaned_reactions_ref = ray.put(cleaned_reactions)

        comp_matrices = {n: [] for n in range(1, max_num_combos + 1)}  # type: ignore
        comp_matrices_refs_dict = {}  # type: ignore

        for n in range(1, max_num_combos + 1):
            comp_matrices_refs_dict[n] = []
            for group in grouper(combinations(range(num_rxns), n), self.chunk_size):
                comp_matrices_refs_dict[n].append(
                    _create_comp_matrices.remote(
                        group, cleaned_reactions_ref, num_entries, net_coeff_filter
                    )
                )

        logger.info("Building comp matrices...")

        num_objs = sum(len(i) for i in comp_matrices_refs_dict.values())  # type: ignore

        with tqdm(total=num_objs) as pbar:
            for n, comp_matrices_refs in comp_matrices_refs_dict.items():
                for comp_matrices_ref in to_iterator(comp_matrices_refs):
                    pbar.update(1)
                    comp_matrices[n].append(comp_matrices_ref)

                comp_matrices[n] = np.concatenate(comp_matrices[n])

                if not comp_matrices[n].any():  # type: ignore
                    del comp_matrices[n]

        logger.info("Comp matrices done...")

        num_cpu_jobs = 0

        c_m_mats = []
        c_m_mats_refs = []

        num_jobs = sum(
            len(val) // self.chunk_size + 1 for val in comp_matrices.values()
        )
        num_batches = int(num_jobs // batch_size + 1)

        batch_count = 1

        for n, comp_matrix in comp_matrices.items():
            if n >= 4:
                num_splits = len(comp_matrix) // self.chunk_size + 1
                splits = np.array_split(comp_matrix, num_splits)
            else:
                splits = [comp_matrix]  # only submit one job for small n

            for group in splits:
                if len(group) == 0:  # catch empty matrices
                    continue

                path_balancer = _balance_path_arrays_cpu_wrapper
                num_cpu_jobs += 1

                c_m_mats_refs.append(
                    path_balancer.remote(
                        group,
                        net_rxn_vector,
                    )
                )

                if len(c_m_mats_refs) >= batch_size:
                    for c_m_mats_ref in tqdm(
                        to_iterator(c_m_mats_refs),
                        total=len(c_m_mats_refs),
                        desc=(
                            f"{self.__class__.__name__} (Batch"
                            f" {batch_count}/{num_batches})"
                        ),
                    ):
                        c_m_mats.append(c_m_mats_ref)

                    batch_count += 1

                    num_cpu_jobs = 0

                    c_m_mats_refs = []

        for c_m_mats_ref in tqdm(
            to_iterator(c_m_mats_refs),
            total=len(c_m_mats_refs),
            desc=f"{self.__class__.__name__} (Batch {batch_count}/{num_batches})",
        ):
            c_m_mats.append(c_m_mats_ref)

        c_mats, m_mats = zip(*c_m_mats)
        c_mats = [mat for mats in c_mats for mat in mats if mat is not None]
        m_mats = [mat for mats in m_mats for mat in mats if mat is not None]

        paths = []
        for c_mat, m_mat in zip(c_mats, m_mats):
            path_rxns = []
            path_costs = []

            for rxn_mat in c_mat:
                ents, coeffs = zip(
                    *[
                        (entries[idx], c)
                        for idx, c in enumerate(rxn_mat)
                        if not np.isclose(c, 0.0)
                    ]
                )

                if self.open_elem is not None:
                    rxn = OpenComputedReaction(
                        entries=ents,
                        coefficients=coeffs,
                        chempots={self.open_elem: self.chempot},
                    )

                else:
                    rxn = ComputedReaction(entries=ents, coefficients=coeffs)

                try:
                    path_rxns.append(rxn)
                    path_costs.append(cleaned_costs[cleaned_reactions.index(rxn)])
                except Exception as e:
                    print(e)
                    continue

            p = BalancedPathway(path_rxns, m_mat.flatten(), path_costs, balanced=True)
            paths.append(p)

        filtered_paths = []
        if filter_interdependent:
            precursor_comps = [p.composition for p in precursors]
            for p in paths:
                interdependent = p.contains_interdependent_rxns(precursor_comps)
                if not interdependent:
                    filtered_paths.append(p)
        else:
            filtered_paths = paths

        filtered_paths = sorted(list(set(filtered_paths)), key=lambda p: p.average_cost)

        return PathwaySet.from_paths(filtered_paths)

    def _find_intermediate_rxns(
        self,
        targets,
        energy_cutoff,
        use_basic_enumerator,
        use_minimize_enumerator,
    ):
        """
        Method for finding intermediate reactions using enumerators and
        specified settings.
        """

        intermediates = {e for rxn in self.reactions for e in rxn.entries}
        intermediates = GibbsEntrySet(
            list(intermediates) + targets,
        )
        target_formulas = [e.composition.reduced_formula for e in targets]
        ref_elems = {e for e in self.entries if e.is_element}

        intermediates = intermediates | ref_elems

        rxn_set = ReactionSet(
            intermediates.entries_list,
            {},
            {},
            open_elem=self.open_elem,
            chempot=self.chempot,
            all_data={},
        )

        if use_basic_enumerator:
            be = BasicEnumerator(targets=target_formulas)
            rxn_set = rxn_set.add_rxn_set(be.enumerate(intermediates))

            if self.open_elem:
                boe = BasicOpenEnumerator(
                    open_phases=[Composition(str(self.open_elem)).reduced_formula],
                    targets=target_formulas,
                )

                rxn_set = rxn_set.add_rxn_set(boe.enumerate(intermediates))

        if use_minimize_enumerator:
            mge = MinimizeGibbsEnumerator(
                targets=target_formulas,
            )
            rxn_set = rxn_set.add_rxn_set(mge.enumerate(intermediates))

            if self.open_elem:
                mgpe = MinimizeGrandPotentialEnumerator(
                    open_elem=self.open_elem,
                    mu=self.chempot,
                    targets=target_formulas,
                )
                rxn_set.add_rxn_set(mgpe.enumerate(intermediates))

        rxns = list(filter(lambda x: x.energy_per_atom < energy_cutoff, rxn_set))
        rxns = [r for r in rxns if all(e in intermediates for e in r.entries)]
        num_rxns = len(rxns)
        rxns = ReactionSet.from_rxns(rxns, filter_duplicates=True)

        logger.info(f"Found {num_rxns} intermediate reactions! \n")

        return rxns

    @property
    def entries(self) -> GibbsEntrySet:
        """Entry set used in solver"""
        return self._entries


@jit(nopython=True)
def _balance_path_arrays_cpu(
    comp_matrices: np.ndarray,
    net_coeffs: np.ndarray,
    tol: float = 1e-6,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Fast solution for reaction multiplicities via mass balance stochiometric
    constraints. Parallelized using Numba JIT. Can be applied to large batches (100K-1M
    sets of reactions at a time.)

    Args:
        comp_matrices: Array containing stoichiometric coefficients of all
            compositions in all reactions, for each trial combination.
        net_coeffs: Array containing stoichiometric coefficients of net reaction.
        tol: numerical tolerance for determining if a multiplicity is zero
            (i.e., if reaction was removed).
    """
    shape = comp_matrices.shape
    net_coeff_filter = np.argwhere(net_coeffs != 0).flatten()
    len_net_coeff_filter = len(net_coeff_filter)
    all_multiplicities = np.zeros((shape[0], shape[1]), np.float64)
    indices = np.full(shape[0], False)

    for i in range(shape[0]):
        correct = True
        for j in range(len_net_coeff_filter):
            idx = net_coeff_filter[j]
            if not comp_matrices[i][:, idx].any():
                correct = False
                break
        if not correct:
            continue

        comp_pinv = np.linalg.pinv(comp_matrices[i]).T
        multiplicities = comp_pinv @ net_coeffs
        solved_coeffs = comp_matrices[i].T @ multiplicities

        if (multiplicities < tol).any():
            continue

        if not (
            np.abs(solved_coeffs - net_coeffs) <= (1e-08 + 1e-05 * np.abs(net_coeffs))
        ).all():
            continue

        all_multiplicities[i] = multiplicities
        indices[i] = True

    filtered_indices = np.argwhere(indices != 0).flatten()
    length = filtered_indices.shape[0]
    filtered_comp_matrices = np.empty((length, shape[1], shape[2]), np.float64)
    filtered_multiplicities = np.empty((length, shape[1]), np.float64)

    for i in range(length):
        idx = filtered_indices[i]
        filtered_comp_matrices[i] = comp_matrices[idx]
        filtered_multiplicities[i] = all_multiplicities[idx]

    return filtered_comp_matrices, filtered_multiplicities


@ray.remote
def _create_comp_matrices(combos, rxns, num_entries, net_coeff_filter):
    """Create array of stoichiometric coefficients for each reaction."""
    comp_matrices = np.stack(
        [
            np.vstack([rxns[r].get_entry_idx_vector(num_entries) for r in combo])
            for combo in combos
            if combo
        ]
    )
    # filter bad matrices
    comp_matrices_filtered = comp_matrices[
        comp_matrices[:, :, net_coeff_filter].any(axis=1).all(axis=1)
    ]

    return comp_matrices_filtered


@ray.remote
def _balance_path_arrays_cpu_wrapper(
    comp_matrices,
    net_rxn_vector,
):
    """Wraps pathway balancing method with ray.remote decorator"""
    return _balance_path_arrays_cpu(comp_matrices, net_rxn_vector)



"""Core definition for various task and synthesis recipe documents"""
from datetime import datetime
from typing import Any, Dict, List  # avoids pydantic issues

from pydantic import BaseModel, Field
from pymatgen.core.composition import Element

from rxn_network.entries.entry_set import GibbsEntrySet
from rxn_network.enumerators.base import Enumerator
from rxn_network.network.base import Network
from rxn_network.pathways.pathway_set import PathwaySet
from rxn_network.pathways.solver import Solver
from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import datetime_str


class EntrySetDocument(BaseModel):
    """
    A single entry set object as produced by the GetEntrySet job.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    entries: GibbsEntrySet = Field(description="The entry set object.")
    e_above_hull: float = Field(None, description="The e_above_hull cutoff.")
    include_polymorphs: bool = Field(
        False, description="Whether to include metastable polymorphs in the entry set."
    )
    formulas_to_include: List[str] = Field(
        None, description="The required formulas to include during construciton."
    )


class EnumeratorTaskDocument(BaseModel):
    """
    A single task object from the reaction enumerator workflow.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    rxns: ReactionSet = Field(description="The reaction set.")
    targets: List[str] = Field(
        None, description="The target formulas used in the enumerator(s)."
    )
    elements: List[Element] = Field(
        None, description="The elements of the total chemical system"
    )
    chemsys: str = Field(
        None, description="The total chemical system string (e.g., Fe-Li-O)."
    )
    added_elements: List[Element] = Field(
        None, description="The elements added beyond the elements of the target(s)."
    )
    added_chemsys: str = Field(
        None, description="The chemical system of the added elements"
    )
    enumerators: List[Enumerator] = Field(
        None,
        description="A list of the enumerator objects used to calculate the reactions.",
    )


class CompetitionTaskDocument(BaseModel):
    """
    A document containing the reactions and their selectivities as created by the
    CalculateCompetitionMaker job.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    rxns: ReactionSet = Field(
        description=(
            "The reaction set, where thereactions have calculated competition"
            " information stored in their data attribute."
        )
    )
    target_formula: str = Field(
        description="The reduced chemical formula of the target material."
    )
    open_elem: Element = Field(None, description="The open element (if any).")
    chempot: float = Field(
        None, description="The chemical potential of the open element."
    )
    added_elements: List[Element] = Field(
        None, description="The elements added beyond the elements of the target(s)."
    )
    added_chemsys: str = Field(
        None, description="The chemical system of the added elements."
    )
    calculate_competition: bool = Field(
        None, description="Whether to calculate competition scores."
    )
    calculate_chempot_distances: bool = Field(
        None, description="Whether to calculate chemical potential distances."
    )
    temp: float = Field(
        None,
        description=(
            "The temperature in K used to determine the primary competition weightings."
        ),
    )
    batch_size: int = Field(None, description="The batch size for the reaction set")
    cpd_kwargs: Dict[str, Any] = Field(
        None, description="The kwargs for ChempotDistanceCalculator."
    )


class NetworkTaskDocument(BaseModel):
    """
    The calculation output from the NetworkMaker workflow.

    Optionally includes unbalanced paths found during pathfinding.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    network: Network = Field(description="The reaction network")
    paths: PathwaySet = Field(
        None, description="The (simple/unbalanced) reaction pathways"
    )
    k: int = Field(None, description="The number of paths solved for, if any.")
    precursors: List[str] = Field(None, description="The precursor formulas.")
    targets: List[str] = Field(None, description="The target formulas.")


class PathwaySolverTaskDocument(BaseModel):
    """
    The calculation output from the PathwaySolverMaker workflow. Contains the pathway
    solver object and its calculated balanced pathways.
    """

    task_label: str = Field(None, description="The name of the task.")
    last_updated: datetime = Field(
        default_factory=datetime_str,
        description="Timestamp of when the document was last updated.",
    )
    solver: Solver = Field(description="The pathway solver used to calculate pathways.")
    balanced_paths: PathwaySet = Field(description="The balanced reaction pathways.")
    precursors: List[str] = Field(description="The precursor compositions.")
    targets: List[str] = Field(description="The target compositions.")
    net_rxn: ComputedReaction = Field(
        description="The net reaction used for pathway solving."
    )
    max_num_combos: int = Field(
        description=(
            "The maximum number of combinations to consider in the pathway solver."
        )
    )
    find_intermediate_rxns: bool = Field(
        description="Whether to find reactions from intermediate compositions."
    )
    intermediate_rxn_energy_cutoff: float = Field(
        description="The mximum energy cutoff for filtering intermediate reactions."
    )
    use_basic_enumerator: bool = Field(
        description="Whether to use the basic enumerators in path solving."
    )
    use_minimize_enumerator: bool = Field(
        description="Whether to use the minimize enumerators in path solving."
    )
    filter_interdependent: bool = Field(
        description="Whether to filter out interdependent pathway.s"
    )



"""
Utility functions used in the thermodynamic analysis classes.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from pymatgen.analysis.phase_diagram import PhaseDiagram
from tqdm import tqdm

if TYPE_CHECKING:
    from pymatgen.entries import Entry


def expand_pd(entries: Iterable[Entry], pbar: bool = False) -> dict[str, PhaseDiagram]:
    """
    Helper method for generating a set of smaller phase diagrams for analyzing
    thermodynamic stability in large chemical systems. This is necessary when
    considering chemical systems which contain 10 or more elements, due to dimensional
    limitations of the Qhull algorithm.

    Args:
        entries ([Entry]): list of Entry objects for building phase diagram.
        pbar (bool): whether to show a progress bar.

    Returns:
        Dictionary of PhaseDiagram objects indexed by chemical subsystem string;
        e.g. {"Li-Mn-O": <PhaseDiagram object>, "C-Y": <PhaseDiagram object>, ...}
    """

    pd_dict: dict[str, PhaseDiagram] = {}

    sorted_entries = sorted(
        entries, key=lambda x: len(x.composition.elements), reverse=True
    )

    for e in tqdm(sorted_entries, disable=not pbar, desc="Building phase diagrams"):
        for chemsys in pd_dict:
            if set(e.composition.chemical_system.split("-")).issubset(
                chemsys.split("-")
            ):
                break
        else:
            pd_dict[e.composition.chemical_system] = PhaseDiagram(
                list(
                    filter(
                        lambda x: set(x.composition.elements).issubset(
                            e.composition.elements
                        ),
                        entries,
                    )
                )
            )

    return pd_dict



"""
Basic interface for a reaction network and its graph.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Iterable

from monty.json import MontyDecoder, MSONable
from rustworkx import PyDiGraph

from rxn_network.entries.entry_set import GibbsEntrySet

if TYPE_CHECKING:
    from pymatgen.entries import Entry

    from rxn_network.costs.base import CostFunction
    from rxn_network.pathways.base import Pathway
    from rxn_network.reactions.reaction_set import ReactionSet


class Network(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction network.
    """

    def __init__(
        self,
        rxns: ReactionSet,
        cost_function: CostFunction,
    ):
        self.rxns = rxns
        self.cost_function = cost_function

        self.entries = GibbsEntrySet(rxns.entries)
        self.entries.build_indices()

        self._precursors = None
        self._target = None
        self._g = None

    @abstractmethod
    def build(self) -> None:
        """Construct the network in-place from the supplied enumerators"""

    @abstractmethod
    def find_pathways(self, target, k) -> list[Pathway]:
        """Find reaction pathways"""

    @abstractmethod
    def set_precursors(self, precursors: Iterable[Entry | str]) -> None:
        """Set the phases used as precursors in the network (in-place)"""

    @abstractmethod
    def set_target(self, target: Entry | str) -> None:
        """Set the phase used as a target in the network (in-place)"""

    def as_dict(self) -> dict:
        """Returns MSONable dict for serialization. See monty package for more
        information."""
        d = super().as_dict()
        d["precursors"] = list(self.precursors) if self.precursors else None
        d["target"] = self.target
        d["graph"] = self.graph.as_dict()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> Network:
        """Instantiate object from MSONable dict. See monty package for more
        information."""
        precursors = d.pop("precursors", None)
        target = d.pop("target", None)
        graph = d.pop("graph", None)

        network = super().from_dict(d)
        network._precursors = precursors  # pylint: disable=protected-access
        network._target = target  # pylint: disable=protected-access
        network._g = MontyDecoder().process_decoded(  # pylint: disable=protected-access
            graph
        )

        return network

    @property
    def precursors(self) -> set[Entry] | None:
        """The phases used as precursors in the network"""
        return self._precursors

    @property
    def target(self) -> Entry | None:
        """The phase used as a target in the network"""
        return self._target

    @property
    def graph(self):
        """Returns the network's Graph object"""
        return self._g

    @property
    def chemsys(self) -> str:
        """A string representing the chemical system (elements) of the network"""
        return "-".join(sorted(self.entries.chemsys))

    def __repr__(self) -> str:
        return (
            "Reaction network for chemical system: "
            f"{self.chemsys}, "
            f"with graph: {str(self.graph)}"
        )

    def __str__(self) -> str:
        return self.__repr__()


class Graph(PyDiGraph):
    """
    Thin wrapper around rx.PyDiGraph to allow for serialization and optimized database
    storage.
    """

    def as_dict(self) -> dict:
        """
        Represents the PyDiGraph object as a serializable dictionary.

        See monty package (MSONable) for more information.
        """
        d = {"@module": self.__class__.__module__, "@class": self.__class__.__name__}

        d["nodes"] = [n.as_dict() for n in self.nodes()]  # type: ignore
        d["node_indices"] = list(self.node_indices())  # type: ignore
        d["edges"] = [
            (*e, obj.as_dict() if hasattr(obj, "as_dict") else obj)  # type: ignore
            for e, obj in zip(self.edge_list(), self.edges())
        ]

        return d

    @classmethod
    def from_dict(cls, d: dict) -> Graph:
        """
        Instantiates a Graph object from a dictionary.

        See as_dict() and monty package (MSONable) for more information.
        """
        nodes = MontyDecoder().process_decoded(d["nodes"])
        node_indices = MontyDecoder().process_decoded(d["node_indices"])
        edges = [(e[0], e[1], MontyDecoder().process_decoded(e[2])) for e in d["edges"]]

        nodes = dict(zip(nodes, node_indices))

        graph = cls()
        new_indices = graph.add_nodes_from(list(nodes.keys()))
        mapping = {nodes[node]: idx for idx, node in zip(new_indices, nodes.keys())}

        new_mapping = []
        for edge in edges:
            new_mapping.append((mapping[edge[0]], mapping[edge[1]], edge[2]))

        graph.add_edges_from(new_mapping)

        return graph

    def __repr__(self) -> str:
        return (
            super().__repr__()
            + f" with {self.num_nodes()} nodes and {self.num_edges()} edges"
        )
